import numpy as np
import pandas as pd
import xarray as xr


@staticmethod
def validate_netcdf_forcing(
    ds: xr.Dataset, required_vars: tuple[str, ...], required_coords: tuple[str, ...]
) -> xr.Dataset:
    """Validate a forcing dataset by checking for required variables and coordinates."""
    # Check variables
    _required_vars = set(required_vars)
    if not _required_vars.issubset(ds.data_vars):
        missing_vars = _required_vars - set(ds.data_vars)
        raise ValueError(
            f"Missing required variables for netcdf forcing: {missing_vars}"
        )

    # Check coordinates
    _required_coords = set(required_coords)
    if not _required_coords.issubset(ds.coords):
        missing_coords = _required_coords - set(ds.coords)
        raise ValueError(
            f"Missing required coordinates for netcdf forcing: {missing_coords}"
        )

    # Check time step
    ts = pd.to_timedelta(np.diff(ds.time).mean())
    if ts < pd.to_timedelta("1H"):
        raise ValueError(
            f"SFINCS NetCDF forcing time step cannot be less than 1 hour: {ts}"
        )

    for var in ds.data_vars:
        # Check order of dimensions
        if ds[var].dims != required_coords:
            raise ValueError(
                f"Order of dimensions for variable {var} must be {required_coords}"
            )
    return ds



import math
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from plotly.express import line
from plotly.express.colors import sample_colorscale
from pydantic import AfterValidator, BaseModel, Field, model_validator
from typing_extensions import Annotated

from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.timeseries import Scstype


def ensure_ascii(s: str):
    assert s.isascii()
    return s


AsciiStr = Annotated[str, AfterValidator(ensure_ascii)]


class Cstype(str, Enum):
    """The accepted input for the variable cstype in Site."""

    projected = "projected"
    spherical = "spherical"


class SCSModel(BaseModel):
    """Class describing the accepted input for the variable scs.

    Includes the file with the non-dimensional SCS rainfall curves in the site folder and the SCS rainfall curve type.

    Attributes
    ----------
    file : str
        The path to the SCS rainfall curves file.
    type : Scstype
        The type of the SCS rainfall curve.
    """

    file: str
    type: Scstype


class RiverModel(BaseModel):
    """Model that describes the accepted input for the variable river in Site.

    Attributes
    ----------
    name : str
        The name of the river.
    description : Optional[str], default=None
        description of the river.
    mean_discharge : us.UnitfulDischarge
        The mean discharge of the river.
    x_coordinate : float
        The x coordinate of the river.
    y_coordinate : float
        The y coordinate of the river.
    """

    name: str
    description: Optional[str] = None
    mean_discharge: us.UnitfulDischarge
    x_coordinate: float
    y_coordinate: float


class ObsPointModel(BaseModel):
    """The accepted input for the variable obs_point in Site.

    obs_points is used to define output locations in the hazard model, which will be plotted in the user interface.

    Attributes
    ----------
    name : Union[int, AsciiStr]
        The name of the observation point.
    description : Optional[str], default=""
        The description of the observation point.
    ID : Optional[int], default=None
        The ID of the observation point.
    file : Optional[str], default=None
        The path to the observation point data file.
    """

    name: Union[int, AsciiStr]
    description: Optional[str] = ""
    ID: Optional[int] = (
        None  # if the observation station is also a tide gauge, this ID should be the same as for obs_station
    )
    file: Optional[str] = None  # for locally stored data
    lat: float
    lon: float


class FloodFrequencyModel(BaseModel):
    """The accepted input for the variable flood_frequency in Site."""

    flooding_threshold: us.UnitfulLength


class DemModel(BaseModel):
    """The accepted input for the variable dem in Site.

    Attributes
    ----------
    filename : str
        The path to the digital elevation model file.
    units : us.UnitTypesLength
        The units of the digital elevation model file.
    """

    filename: str
    units: us.UnitTypesLength


class DatumModel(BaseModel):
    """
    The accepted input for the variable datums in WaterlevelReferenceModel.

    Attributes
    ----------
    name : str
        The name of the vertical reference model.
    height : us.UnitfulLength
        The height of the vertical reference model relative to the main reference.
    """

    name: str
    height: us.UnitfulLength


class WaterlevelReferenceModel(BaseModel):
    """The accepted input for the variable water_level in Site.

    Waterlevels timeseries are calculated from user input, assumed to be relative to the `reference` vertical reference model.

    For plotting in the GUI, the `reference` vertical reference model is used as the main zero-reference, all values are relative to this.
    All other vertical reference models are plotted as dashed lines.

    Attributes
    ----------
    reference : str
        The name of the vertical reference model that is used as the main zero-reference.
    datums : list[DatumModel]
        The vertical reference models that are used to calculate the waterlevels timeseries.
        The datums are used to calculate the waterlevels timeseries, which are relative to the `reference` vertical reference model.
    """

    reference: str
    datums: list[DatumModel] = Field(default_factory=list)

    def get_datum(self, name: str) -> DatumModel:
        for datum in self.datums:
            if datum.name == name:
                return datum
        raise ValueError(f"Could not find datum with name {name}")

    @model_validator(mode="after")
    def main_reference_should_be_in_datums_and_eq_zero(self):
        if self.reference not in [datum.name for datum in self.datums]:
            raise ValueError(f"Reference {self.reference} not in {self.datums}")
        if not math.isclose(
            self.get_datum(self.reference).height.value, 0, abs_tol=1e-6
        ):
            raise ValueError(f"Reference {self.reference} height is not zero")
        return self

    @model_validator(mode="after")
    def all_datums_should_have_unique_names(self):
        datum_names = [datum.name for datum in self.datums]
        if len(set(datum_names)) != len(datum_names):
            raise ValueError(f"Duplicate datum names found: {datum_names}")
        return self


class CycloneTrackDatabaseModel(BaseModel):
    """The accepted input for the variable cyclone_track_database in Site.

    Attributes
    ----------
    file : str
        The path to the cyclone track database file.
    """

    file: str


class SlrScenariosModel(BaseModel):
    """The accepted input for the variable slr_scenarios.

    Attributes
    ----------
    file : str
        The path to the sea level rise scenarios file.
    relative_to_year : int
        The year to which the sea level rise scenarios are relative.
    """

    file: str
    relative_to_year: int

    def interp_slr(
        self,
        scenario: str,
        year: float,
        units: us.UnitTypesLength = us.UnitTypesLength.meters,
    ) -> float:
        """Interpolate SLR value and reference it to the SLR reference year from the site toml.

        Parameters
        ----------
        scenario : str
            SLR scenario name to use from the column names in self.file
        year : float
            year to evaluate
        units : us.UnitTypesLength, default = us.UnitTypesLength.meters
            The units to convert the SLR value to. Default is meters.

        Returns
        -------
        float
            The interpolated sea level rise value in the specified units, relative to the reference year.

        Raises
        ------
        ValueError
            if the reference year is outside of the time range in the slr.csv file
        ValueError
            if the year to evaluate is outside of the time range in the slr.csv file
        """
        df = pd.read_csv(self.file)
        if year > df["year"].max() or year < df["year"].min():
            raise ValueError(
                "The selected year is outside the range of the available SLR scenarios"
            )

        if (
            self.relative_to_year > df["year"].max()
            or self.relative_to_year < df["year"].min()
        ):
            raise ValueError(
                f"The reference year {self.relative_to_year} is outside the range of the available SLR scenarios"
            )

        slr = np.interp(year, df["year"], df[scenario])
        ref_slr = np.interp(self.relative_to_year, df["year"], df[scenario])

        new_slr = us.UnitfulLength(
            value=slr - ref_slr,
            units=df["units"][0],
        )
        return np.round(new_slr.convert(units), decimals=2)

    def plot_slr_scenarios(
        self,
        scenario_names: list[str],
        output_loc: Path,
        units: us.UnitTypesLength = us.UnitTypesLength.meters,
    ) -> str:
        """
        Plot sea level rise scenarios.

        Returns
        -------
        html_path : str
            The path to the html plot of the sea level rise scenarios.
        """
        df = pd.read_csv(self.file)

        ncolors = len(df.columns) - 2
        if "units" not in df.columns:
            raise ValueError(f"Expected column `units` in {self.file}.")

        _units = df["units"].iloc[0]
        _units = us.UnitTypesLength(_units)

        if "year" not in df.columns:
            raise ValueError(f"Expected column `year` in {self.file}.")

        if (
            self.relative_to_year > df["year"].max()
            or self.relative_to_year < df["year"].min()
        ):
            raise ValueError(
                f"The reference year {self.relative_to_year} is outside the range of the available SLR scenarios"
            )

        for scn in scenario_names:
            ref_slr = np.interp(self.relative_to_year, df["year"], df[scn])
            df[scn] -= ref_slr

        df = df.drop(columns="units").melt(id_vars=["year"]).reset_index(drop=True)
        # convert to units used in GUI
        conversion_factor = us.UnitfulLength(value=1.0, units=_units).convert(units)
        df.iloc[:, -1] = (conversion_factor * df.iloc[:, -1]).round(decimals=2)

        # rename column names that will be shown in html
        df = df.rename(
            columns={
                "variable": "Scenario",
                "value": f"Sea level rise [{units.value}]",
            }
        )

        colors = sample_colorscale(
            "rainbow", [n / (ncolors - 1) for n in range(ncolors)]
        )
        fig = line(
            df,
            x="year",
            y=f"Sea level rise [{units.value}]",
            color="Scenario",
            color_discrete_sequence=colors,
        )

        # fig.update_traces(marker={"line": {"color": "#000000", "width": 2}})

        fig.update_layout(
            autosize=False,
            height=100 * 1.2,
            width=280 * 1.3,
            margin={"r": 0, "l": 0, "b": 0, "t": 0},
            font={"size": 10, "color": "black", "family": "Arial"},
            title_font={"size": 10, "color": "black", "family": "Arial"},
            legend_font={"size": 10, "color": "black", "family": "Arial"},
            legend_grouptitlefont={"size": 10, "color": "black", "family": "Arial"},
            legend={"entrywidthmode": "fraction", "entrywidth": 0.2},
            yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
            xaxis_title=None,
            xaxis_range=[self.relative_to_year, df["year"].max()],
            legend_title=None,
            # paper_bgcolor="#3A3A3A",
            # plot_bgcolor="#131313",
        )

        # write html to results folder
        output_loc.parent.mkdir(parents=True, exist_ok=True)
        fig.write_html(output_loc)
        return output_loc.as_posix()


class FloodModel(BaseModel):
    """The accepted input for the variable overland_model and offshore_model in Site.

    Attributes
    ----------
    name : str
        The name of the directory in `static/templates/<directory>` that contains the template model files.
    reference : str
        The name of the vertical reference model that is used as the reference datum. Should be defined in water_level.datums.
    vertical_offset : Optional[us.UnitfulLength], default = None
        The vertical offset of the vertical reference model relative to the main reference.
        Given that the height of the vertical reference model is often determined by external sources,
        this vertical offset can be used to correct systematic over-/underestimation of a vertical reference model.
    """

    name: str
    reference: str

    # this used to be water_level_offset from events
    vertical_offset: Optional[us.UnitfulLength] = None



from flood_adapt.objects.benefits.benefits import Benefit, CurrentSituationModel
from flood_adapt.objects.events.event_factory import EventFactory
from flood_adapt.objects.events.event_set import EventSet, SubEventModel
from flood_adapt.objects.events.events import (
    Event,
    Mode,
    Template,
)
from flood_adapt.objects.events.historical import HistoricalEvent
from flood_adapt.objects.events.hurricane import HurricaneEvent
from flood_adapt.objects.events.synthetic import SyntheticEvent
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    ForcingType,
    IForcing,
)
from flood_adapt.objects.forcing.forcing_factory import ForcingFactory
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.measures.measure_factory import MeasureFactory
from flood_adapt.objects.measures.measures import (
    Buyout,
    Elevate,
    FloodProof,
    FloodWall,
    GreenInfrastructure,
    Measure,
    MeasureType,
    Pump,
    SelectionType,
)
from flood_adapt.objects.object_model import Object
from flood_adapt.objects.projections.projections import (
    PhysicalProjection,
    Projection,
    SocioEconomicChange,
)
from flood_adapt.objects.scenarios.scenarios import Scenario
from flood_adapt.objects.strategies.strategies import Strategy

__all__ = [
    # Object
    "Object",
    # Measures
    "MeasureFactory",
    "Measure",
    "MeasureType",
    "SelectionType",
    "Buyout",
    "Elevate",
    "FloodProof",
    "FloodWall",
    "GreenInfrastructure",
    "Pump",
    # Events
    "Event",
    "EventFactory",
    "SyntheticEvent",
    "HistoricalEvent",
    "HurricaneEvent",
    "Mode",
    "Template",
    # EventSet
    "EventSet",
    "SubEventModel",
    # Forcing
    "ForcingFactory",
    "TimeFrame",
    "IForcing",
    "ForcingType",
    "ForcingSource",
    # Projections
    "Projection",
    "PhysicalProjection",
    "SocioEconomicChange",
    # Strategies
    "Strategy",
    # Scenarios
    "Scenario",
    # Benefits
    "Benefit",
    "CurrentSituationModel",
]



import logging
import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import List, Optional


class FloodAdaptLogging:
    _DEFAULT_FORMATTER = logging.Formatter(
        fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %I:%M:%S %p",
    )
    _root_logger = logging.getLogger("FloodAdapt")

    def __init__(
        self,
        file_path: Optional[Path] = None,
        level: int = logging.INFO,
        formatter: logging.Formatter = _DEFAULT_FORMATTER,
        ignore_warnings: Optional[list[type[Warning]]] = None,
    ) -> None:
        """Initialize the logging system for the FloodAdapt."""
        self._formatter = formatter

        self._root_logger.setLevel(level)
        if self._root_logger.hasHandlers():
            self._root_logger.handlers.clear()

        # Add file handler if provided
        if file_path is not None:
            self.add_file_handler(file_path, level, formatter)

        # Add console handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_handler.setFormatter(formatter)
        self._root_logger.addHandler(console_handler)

        if ignore_warnings:
            for warn_type in ignore_warnings:
                self.configure_warnings("ignore", category=warn_type)

    @classmethod
    def add_file_handler(
        cls,
        file_path: Path,
        level: int = logging.DEBUG,
        formatter: Optional[logging.Formatter] = None,
    ) -> None:
        """Add a file handler to the logger that directs outputs to a the file."""
        if not file_path:
            raise ValueError("file_path must be provided.")
        file_path = Path(file_path)

        # check if the path is a only a filename
        if not file_path.parents:
            file_path = Path.cwd() / file_path
        file_path.parent.mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(filename=file_path, mode="a")
        file_handler.setLevel(level)

        formatter = formatter or cls._DEFAULT_FORMATTER
        file_handler.setFormatter(formatter)

        cls.getLogger().addHandler(file_handler)

    @classmethod
    def remove_file_handler(cls, file_path: Path) -> None:
        """Remove a file handler from the logger, which stops sending logs to that file and closes it."""
        for handler in cls.getLogger().handlers:
            if isinstance(handler, logging.FileHandler) and handler.baseFilename == str(
                file_path.resolve()
            ):
                handler.close()
                cls.getLogger().removeHandler(handler)

    @classmethod
    def getLogger(
        cls, name: Optional[str] = None, level: Optional[int] = None
    ) -> logging.Logger:
        """Get a logger with the specified name. If no name is provided, return the root logger.

        If the logger does not exist, it is created with the specified level.

        Parameters
        ----------
        name : str, optional
            The name of the logger. If not provided, the root logger is returned.
        level : int,
            The level of the logger. The default is logging.INFO.

        Returns
        -------
        logging.Logger
            The logger with the specified name.
        """
        if name is None:
            logger = cls._root_logger
        else:
            logger = logging.getLogger(f"FloodAdapt.{name}")

        if level:
            logger.setLevel(level)

        return logger

    @classmethod
    def set_global_level(cls, level: int) -> None:
        """Set the logging level for FloodAdapt."""
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        for logger in loggers:
            logger.setLevel(level)

    @classmethod
    def shutdown(cls):
        root_logger = cls.getLogger()
        handlers = root_logger.handlers[:]
        for handler in handlers:
            if isinstance(handler, logging.FileHandler):
                handler.close()
                root_logger.removeHandler(handler)
        logging.shutdown()

    @classmethod
    @contextmanager
    def to_file(
        cls,
        *,
        file_path: Path,
        level: int = logging.DEBUG,
        formatter: logging.Formatter = _DEFAULT_FORMATTER,
    ):
        """Open a file at filepath to write logs to. Does not affect other loggers.

        When the context manager exits (via regular execution or an exception), the file is closed and the handler is removed.
        """
        if file_path is None:
            raise ValueError(
                "file_path must be provided as a key value pair: 'file_path=<file_path>'."
            )
        cls.add_file_handler(file_path, level, formatter)
        try:
            yield
        finally:
            cls.remove_file_handler(file_path)

    @classmethod
    def _close_file_handlers(cls, logger: logging.Logger, exclude: List[Path]) -> None:
        """Close and remove file handlers from a logger."""
        for handler in logger.handlers[:]:
            if (
                isinstance(handler, logging.FileHandler)
                and Path(handler.baseFilename) not in exclude
            ):
                handler.close()
                logger.removeHandler(handler)

    @classmethod
    def close_files(cls, exclude: List[Path] = []) -> None:
        """Close all file handlers except those in the exclude list."""
        cls._close_file_handlers(cls.getLogger(), exclude)
        cls._close_file_handlers(cls.getLogger("hydromt"), exclude)

    @classmethod
    def deprecation_warning(cls, version: str, reason: str):
        """Log a deprecation warning with reason and the version that will remove it."""
        warnings.warn(
            f"DeprecationWarning: {reason}. This will be removed in version {version}.",
            DeprecationWarning,
            stacklevel=2,
        )

    @classmethod
    def configure_warnings(
        cls, action: str = "default", category: Optional[type[Warning]] = None
    ):
        """
        Configure the behavior of Python warnings.

        Parameters
        ----------
        action : str, optional
            The action to take on warnings. Common actions include 'ignore', 'default', 'error', 'always', etc.
            The default is 'default', which shows warnings once per triggering location.
        category : type[Warning], optional
            The category of warnings to configure. If not provided, all warnings are configured.
            categories include DeprecationWarning, UserWarning, RuntimeWarning, etc.

        """
        warnings.simplefilter(action=action, category=category)



from datetime import datetime
from pathlib import Path

import cht_meteo
import numpy as np
import xarray as xr
from cht_meteo.dataset import MeteoDataset

from flood_adapt.objects.forcing.time_frame import TimeFrame


class MeteoHandler:
    def __init__(self, dir: Path, lat: float, lon: float) -> None:
        self.dir: Path = dir
        self.dir.mkdir(parents=True, exist_ok=True)

        # Create GFS dataset
        self.dataset = cht_meteo.dataset(
            name="gfs_anl_0p50",
            source="gfs_analysis_0p50",
            path=self.dir,
            lon_range=(lon - 10, lon + 10),
            lat_range=(lat - 10, lat + 10),
        )
        # quick fix for sites near the 0 degree longitude -> shift the meteo download area either east or west of the 0 degree longitude
        # TODO implement a good solution to this in cht_meteo
        self.dataset.lon_range = self._shift_grid_to_positive_lon(self.dataset)

    def download(self, time: TimeFrame):
        # Download and collect data
        time_range = self.get_time_range(time)

        self.dataset.download(time_range=time_range)

    def read(self, time: TimeFrame) -> xr.Dataset:
        self.download(time)
        time_range = self.get_time_range(time)
        ds = self.dataset.collect(time_range=time_range)

        if ds is None:
            raise FileNotFoundError(
                f"No meteo files found in meteo directory {self.dir}"
            )

        ds.raster.set_crs(4326)

        # Rename the variables to match what hydromt-sfincs expects
        ds = ds.rename(
            {
                "barometric_pressure": "press_msl",
                "precipitation": "precip",
                "wind_u": "wind10_u",
                "wind_v": "wind10_v",
            }
        )

        # Convert the longitude to -180 to 180 to match hydromt-sfincs
        if ds["lon"].min() > 180:
            ds["lon"] = ds["lon"] - 360

        return ds

    @staticmethod
    def get_time_range(time: TimeFrame) -> tuple:
        t0 = time.start_time
        t1 = time.end_time
        if not isinstance(t0, datetime):
            t0 = datetime.strptime(t0, "%Y%m%d %H%M%S")
        if not isinstance(t1, datetime):
            t1 = datetime.strptime(t1, "%Y%m%d %H%M%S")
        time_range = (t0, t1)
        return time_range

    @staticmethod
    def _shift_grid_to_positive_lon(grid: MeteoDataset):
        """Shift the grid to positive longitudes if the grid crosses the 0 degree longitude."""
        if np.prod(grid.lon_range) < 0:
            if np.abs(grid.lon_range[0]) > np.abs(grid.lon_range[1]):
                grid.lon_range = [
                    grid.lon_range[0] - grid.lon_range[1] - 1,
                    -1,
                ]
            else:
                grid.lon_range = [
                    1,
                    grid.lon_range[1] - grid.lon_range[0] + 1,
                ]
        return grid.lon_range



from pathlib import Path

from pydantic import BaseModel

from flood_adapt.config.impacts import FloodmapType
from flood_adapt.objects.events.events import Mode


class FloodMap(BaseModel):
    name: str
    map_type: FloodmapType
    mode: Mode
    paths: list[Path]



import os
from pathlib import Path
from typing import Annotated, Any

import pandas as pd
import xarray as xr

from flood_adapt.misc.utils import (
    copy_file_to_output_dir,
    validate_file_extension,
)
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    IWind,
)
from flood_adapt.objects.forcing.netcdf import validate_netcdf_forcing
from flood_adapt.objects.forcing.timeseries import (
    CSVTimeseries,
    SyntheticTimeseries,
    TimeFrame,
    TimeseriesFactory,
)


class WindConstant(IWind):
    source: ForcingSource = ForcingSource.CONSTANT

    speed: us.UnitfulVelocity
    direction: us.UnitfulDirection

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        time = pd.date_range(
            start=time_frame.start_time,
            end=time_frame.end_time,
            freq=time_frame.time_step,
            name="time",
        )
        data = {
            "mag": [self.speed.value for _ in range(len(time))],
            "dir": [self.direction.value for _ in range(len(time))],
        }
        return pd.DataFrame(data=data, index=time)


class WindSynthetic(IWind):
    source: ForcingSource = ForcingSource.SYNTHETIC

    magnitude: SyntheticTimeseries
    direction: SyntheticTimeseries

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        time = pd.date_range(
            start=time_frame.start_time,
            end=time_frame.end_time,
            freq=time_frame.time_step,
            name="time",
        )

        magnitude = TimeseriesFactory.from_object(self.magnitude).to_dataframe(
            time_frame=time_frame,
        )

        direction = TimeseriesFactory.from_object(self.direction).to_dataframe(
            time_frame=time_frame,
        )
        return pd.DataFrame(
            index=time,
            data={
                "mag": magnitude.reindex(time).to_numpy().flatten(),
                "dir": direction.reindex(time).to_numpy().flatten(),
            },
        )


class WindTrack(IWind):
    source: ForcingSource = ForcingSource.TRACK

    path: Annotated[Path, validate_file_extension([".cyc", ".spw"])]
    # path to cyc file, set this when creating it

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        if self.path:
            self.path = copy_file_to_output_dir(self.path, Path(output_dir))


class WindCSV(IWind):
    source: ForcingSource = ForcingSource.CSV

    path: Annotated[Path, validate_file_extension([".csv"])]

    units: dict[str, Any] = {
        "speed": us.UnitTypesVelocity.mps,
        "direction": us.UnitTypesDirection.degrees,
    }

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        return CSVTimeseries.load_file(
            path=self.path, units=us.UnitfulVelocity(value=0, units=self.units["speed"])
        ).to_dataframe(time_frame)

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        self.path = copy_file_to_output_dir(self.path, Path(output_dir))


class WindMeteo(IWind):
    source: ForcingSource = ForcingSource.METEO


class WindNetCDF(IWind):
    source: ForcingSource = ForcingSource.NETCDF
    units: us.UnitTypesVelocity = us.UnitTypesVelocity.mps

    path: Annotated[Path, validate_file_extension([".nc"])]

    def read(self) -> xr.Dataset:
        required_vars = ("wind10_v", "wind10_u", "press_msl")
        required_coords = ("time", "lat", "lon")
        with xr.open_dataset(self.path) as ds:
            validated_ds = validate_netcdf_forcing(ds, required_vars, required_coords)
        return validated_ds

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        self.path = copy_file_to_output_dir(self.path, Path(output_dir))



from os import environ, listdir
from pathlib import Path
from typing import Optional

import tomli
import tomli_w
from pydantic import (
    Field,
    computed_field,
    field_serializer,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """
    The configuration settings for the FloodAdapt database and integrator.

    Precedence is as follows: user arguments > environment variables > defaults in this class.
    When loading is done, the settings are validated and the environment variables are updated with the loaded values.

    If any required settings are missing or invalid, a ValidationError is raised.

    Usage
    -----
    from flood_adapt import Settings

    One of the following:

    1) Load settings from environment variables, if no environment variables are set, use defaults defined in the class:
        `settings = Settings()`

    2) Load settings from a .toml file, overwriting any environment variables set:
        `settings = Settings.read(toml_path: Path)`

    3) Load settings from keyword arguments, overwriting any environment variables:
        `settings = Settings(DATABASE_ROOT="path/to/database", DATABASE_NAME="database_name")`

    Attributes
    ----------
    database_name : str
        The name of the database. Alias: `DATABASE_NAME` (environment variable).
    database_root : Path
        The root directory of the database. Alias: `DATABASE_ROOT` (environment variable).
    delete_crashed_runs : bool
        Whether to delete crashed/corrupted runs immediately after they are detected. Alias: `DELETE_CRASHED_RUNS` (environment variable).
    validate_allowed_forcings : bool
        Whether to validate the forcing types and sources against the allowed forcings in the event model. Alias: `VALIDATE_ALLOWED_FORCINGS` (environment variable).
    validate_binaries : bool
        Whether to validate the existence of the paths to the SFINCS and FIAT binaries. Alias: `VALIDATE_BINARIES` (environment variable).
    sfincs_bin_path : Path
        The path to the SFINCS binary. Alias: `SFINCS_BIN_PATH` (environment variable).
    fiat_bin_path : Path
        The path to the FIAT binary. Alias: `FIAT_BIN_PATH` (environment variable).

    Properties
    ----------
    database_path : Path
        The full path to the database.

    Raises
    ------
    ValidationError
        If required settings are missing or invalid.
    """

    model_config = SettingsConfigDict(env_ignore_empty=True, validate_default=True)

    database_root: Path = Field(
        alias="DATABASE_ROOT",  # environment variable DATABASE_ROOT
        default=Path(__file__).parents[3]
        / "Database",  # If you clone FloodAdapt, default is to look for the Database next to the FloodAdapt folder
        description="The root directory of the database that contains site(s). Usually the directory name is 'Database'. Default is to look for the Database in the same dir as the FloodAdapt cloned repo.",
    )
    database_name: str = Field(
        alias="DATABASE_NAME",  # environment variable DATABASE_NAME
        default="",
        description="The name of the database site, should be a folder inside the database root. The site must contain an 'input' and 'static' folder.",
    )

    delete_crashed_runs: bool = Field(
        alias="DELETE_CRASHED_RUNS",  # environment variable: DELETE_CRASHED_RUNS
        default=False,
        description="Whether to delete the output of crashed/corrupted runs. Be careful when setting this to False, as it may lead to a broken database that cannot be read in anymore.",
        exclude=True,
    )
    validate_allowed_forcings: bool = Field(
        alias="VALIDATE_ALLOWED_FORCINGS",  # environment variable: VALIDATE_ALLOWED_FORCINGS
        default=False,
        description="Whether to validate the forcing types and sources against the allowed forcings in the event model.",
        exclude=True,
    )
    validate_binaries: bool = Field(
        alias="VALIDATE_BINARIES",  # environment variable: VALIDATE_BINARIES
        default=False,
        description="Whether to validate the existence of the paths to the SFINCS and FIAT binaries.",
        exclude=True,
    )

    sfincs_bin_path: Optional[Path] = Field(
        default=None,
        alias="SFINCS_BIN_PATH",  # environment variable: SFINCS_BIN_PATH
        description="The path of the sfincs binary.",
        exclude=True,
    )

    fiat_bin_path: Optional[Path] = Field(
        default=None,
        alias="FIAT_BIN_PATH",  # environment variable: FIAT_BIN_PATH
        description="The path of the fiat binary.",
        exclude=True,
    )

    @computed_field
    @property
    def database_path(self) -> Path:
        return self.database_root / self.database_name

    @model_validator(mode="after")
    def validate_settings(self):
        self._validate_database_path()
        if self.validate_binaries:
            self._validate_fiat_path()
            self._validate_sfincs_path()
        self._update_environment_variables()
        return self

    def _update_environment_variables(self):
        environ["DATABASE_ROOT"] = str(self.database_root)
        environ["DATABASE_NAME"] = self.database_name

        if self.delete_crashed_runs:
            environ["DELETE_CRASHED_RUNS"] = str(self.delete_crashed_runs)
        else:
            environ.pop("DELETE_CRASHED_RUNS", None)

        if self.validate_allowed_forcings:
            environ["VALIDATE_ALLOWED_FORCINGS"] = str(self.validate_allowed_forcings)
        else:
            environ.pop("VALIDATE_ALLOWED_FORCINGS", None)

        if self.validate_binaries:
            environ["VALIDATE_BINARIES"] = str(self.validate_binaries)
            environ["SFINCS_BIN_PATH"] = str(self.sfincs_bin_path)
            environ["FIAT_BIN_PATH"] = str(self.fiat_bin_path)
        else:
            environ.pop("VALIDATE_BINARIES", None)

        return self

    def _validate_database_path(self):
        if not self.database_root.is_dir():
            raise ValueError(f"Database root {self.database_root} does not exist.")

        if self.database_name == "":
            # If database_name is not given as arg or set in env, compute default as the first dir in database_root excluding 'system'
            sites = [
                d
                for d in listdir(self.database_root)
                if d != "system" and not d.startswith(".")
            ]
            if not sites:
                raise ValueError(f"No databases found in {self.database_root}.")
            self.database_name = sites[0]

        if not self.database_path.is_dir():
            raise ValueError(
                f"Database {self.database_name} at {self.database_root} does not exist. Full path: {self.database_path}"
            )

        if not (self.database_path / "input").is_dir():
            raise ValueError(
                f"Database {self.database_name} at {self.database_path} does not contain an input folder."
            )

        if not (self.database_path / "static").is_dir():
            raise ValueError(
                f"Database {self.database_name} at {self.database_path} does not contain a static folder."
            )

        return self

    def _validate_sfincs_path(self):
        if not self.sfincs_bin_path.exists():
            raise ValueError(f"SFINCS binary {self.sfincs_bin_path} does not exist.")
        return self

    def _validate_fiat_path(self):
        if not self.fiat_bin_path.exists():
            raise ValueError(f"FIAT binary {self.fiat_bin_path} does not exist.")
        return self

    @field_serializer("database_root", "database_path")
    def serialize_path(self, path: Path) -> str:
        return str(path)

    @staticmethod
    def read(toml_path: Path) -> "Settings":
        """
        Parse the configuration file and return the parsed settings.

        Parameters
        ----------
        toml_path : Path
            The path to the configuration file.

        Returns
        -------
        Settings
            The parsed configuration settings.

        Raises
        ------
        ValidationError
            If required configuration values are missing or if there is an error parsing the configuration file.
        """
        with open(toml_path, "rb") as f:
            settings = tomli.load(f)

        return Settings(**settings)

    def write(self, toml_path: Path) -> None:
        """
        Write the configuration settings to a .toml file.

        Parameters
        ----------
        toml_path : Path
            The path to the configuration file.

        Returns
        -------
        None

        """
        toml_path = Path(toml_path).resolve()
        if not toml_path.parent.exists():
            toml_path.parent.mkdir(parents=True, exist_ok=True)

        with open(toml_path, "wb") as f:
            tomli_w.dump(
                self.model_dump(
                    by_alias=True,
                    exclude={"sfincs_bin_path", "fiat_bin_path", "database_path"},
                ),
                f,
            )



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class FloodmapType(str, Enum):
    """The accepted input for the variable floodmap in Site."""

    water_level = "water_level"
    water_depth = "water_depth"


class BenefitsModel(BaseModel):
    """The accepted input for the variable benefits in Site.

    Attributes
    ----------
    current_year : int
        The current year used in benefits calculations.
    current_projection : str
        The current projection used in benefits calculations.
    baseline_strategy : str
        The baseline strategy used in benefits calculations.
    event_set : str
        The event set used in benefits calculations.
    """

    current_year: int
    current_projection: str
    baseline_strategy: str
    event_set: str


class EquityModel(BaseModel):
    """
    The accepted input for the variable equity in Site.

    Attributes
    ----------
    census_data : str
        TODO
    percapitaincome_label : Optional[str], default="PerCapitaIncome"
        TODO
    totalpopulation_label : Optional[str], default="TotalPopulation"
        TODO
    """

    census_data: str
    percapitaincome_label: Optional[str] = "PerCapitaIncome"
    totalpopulation_label: Optional[str] = "TotalPopulation"


class AggregationModel(BaseModel):
    """The accepted input for the variable aggregation in Site.

    Attributes
    ----------
    name : str
        TODO
    file : str
        TODO
    field_name : str
        TODO
    equity : Optional[EquityModel], default=None
        TODO
    """

    name: str
    file: str
    field_name: str
    equity: Optional[EquityModel] = None


class BFEModel(BaseModel):
    """The accepted input for the variable bfe in Site.

    Attributes
    ----------
    geom : str
        TODO
    table : Optional[str], default=None
        TODO
    field_name : str
        TODO
    """

    geom: str
    table: Optional[str] = None
    field_name: str


class SVIModel(BaseModel):
    """The accepted input for the variable svi in Site.

    Attributes
    ----------
    geom : str
        TODO
    field_name : str
        TODO
    """

    geom: str
    field_name: str


class NoFootprintsModel(BaseModel):
    """
    The configuration on the how to show objects with no footprints.

    Attributes
    ----------
    shape : Optional[str], default="triangle"
        The shape of the object with no footprints.
    diameter_meters : Optional[float], default=10
        The diameter of the object with no footprints in meters.
    """

    shape: Optional[str] = "triangle"
    diameter_meters: Optional[float] = 10


class RiskModel(BaseModel):
    """The accepted input for the variable risk in Site.

    Attributes
    ----------
    return_periods : list[int]
        The return periods for the risk model.
    """

    return_periods: list = [1, 2, 5, 10, 25, 50, 100]


class DamageType(str, Enum):
    """The accepted input for the variable footprints_dmg_type."""

    absolute = "absolute"
    relative = "relative"



import re
from enum import Enum
from pathlib import Path
from typing import Any, Literal, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import tomli
from pydantic import BaseModel, Field, model_validator

from flood_adapt.config.impacts import DamageType
from flood_adapt.objects.forcing import unit_system as us


class Layer(BaseModel):
    """
    Base class for layers in the GUI.

    Attributes
    ----------
    bins : list[float]
        The bins for the layer.
    colors : list[str]
        The colors for the layer.
    """

    bins: list[float]
    colors: list[str]
    decimals: Optional[int] = 0

    @model_validator(mode="after")
    def check_bins_and_colors(self) -> "Layer":
        """Check that the bins and colors have the same length."""
        if (len(self.bins) + 1) != len(self.colors):
            raise ValueError(
                f"Number of bins ({len(self.bins)}) must be one less than number of colors ({len(self.colors)})"
            )
        return self


class FloodMapLayer(Layer):
    zbmax: float
    depth_min: float
    roads_min_zoom_level: int = 14


class FootprintsDmgLayer(Layer):
    type: DamageType = DamageType.absolute


class BenefitsLayer(Layer):
    threshold: Optional[float] = None


class LogicalOperator(str, Enum):
    AND = "and"
    OR = "or"


class FieldName(str, Enum):
    """Enum for valid field names with mapping to dictionary keys."""

    NAME = "name"
    LONG_NAME = "long_name"
    DESCRIPTION = "description"

    @property
    def dict_key(self) -> str:
        """Get the actual dictionary key for this field name."""
        mapping = {
            "name": "name",
            "long_name": "Long Name",
            "description": "Description",
        }
        return mapping[self.value]


class FilterCondition(BaseModel):
    """A single filter condition."""

    field_name: FieldName
    values: list[Any]
    operator: LogicalOperator = (
        LogicalOperator.OR
    )  # How to combine values within this condition


class FilterGroup(BaseModel):
    """A group of filter conditions with logical operators."""

    conditions: list[FilterCondition]
    operator: LogicalOperator = (
        LogicalOperator.OR
    )  # How to combine conditions within this group


class MetricLayer(Layer):
    type: str
    # Simplified: just a single FilterGroup or FilterCondition
    filters: Union[FilterGroup, FilterCondition] = Field(
        default_factory=lambda: FilterGroup(conditions=[])
    )

    def matches(self, data_dict: dict) -> bool:
        """Check if the given data dictionary matches the filter criteria."""
        if isinstance(self.filters, FilterCondition):
            return self._evaluate_condition(self.filters, data_dict)
        else:  # FilterGroup
            return self._evaluate_filter_group(self.filters, data_dict)

    def _evaluate_filter_group(self, group: FilterGroup, data_dict: dict) -> bool:
        """Evaluate a single filter group."""
        if not group.conditions:
            return True

        condition_results = []
        for condition in group.conditions:
            condition_results.append(self._evaluate_condition(condition, data_dict))

        if group.operator == LogicalOperator.AND:
            return all(condition_results)
        else:  # OR
            return any(condition_results)

    def _evaluate_condition(self, condition: FilterCondition, data_dict: dict) -> bool:
        """Evaluate a single condition."""
        # Use the dict_key property to get the actual dictionary key
        field_value = data_dict.get(condition.field_name.dict_key)
        if field_value is None:
            return False

        value_matches = [value in field_value for value in condition.values]  # noqa: PD011

        if condition.operator == LogicalOperator.AND:
            return all(value_matches)
        else:  # OR
            return any(value_matches)


class AggregationDmgLayer(MetricLayer):
    type: str = "damage"
    filters: FilterGroup = Field(
        default_factory=lambda: FilterGroup(
            conditions=[
                FilterCondition(
                    field_name=FieldName.NAME,
                    values=[
                        "TotalDamageEvent",
                        "ExpectedAnnualDamages",
                        "TotalDamageRP",
                        "EWEAD",
                    ],
                )
            ]
        )
    )


class OutputLayers(BaseModel):
    """The configuration of the mapbox layers in the gui.

    Attributes
    ----------
    floodmap : FloodMapLayer
        The configuration of the floodmap layer.
    aggregation_dmg : AggregationDmgLayer
        The configuration of the aggregation damage layer.
    footprints_dmg : FootprintsDmgLayer
        The configuration of the footprints damage layer.

    benefits : BenefitsLayer
        The configuration of the benefits layer.
    """

    floodmap: FloodMapLayer
    aggregation_dmg: AggregationDmgLayer
    footprints_dmg: FootprintsDmgLayer
    aggregated_metrics: list[MetricLayer] = Field(default_factory=list)
    benefits: Optional[BenefitsLayer] = None

    def get_aggr_metrics_layers(
        self,
        metrics: list[dict],
        type: Literal["single_event", "risk"] = "single_event",
        rp: Optional[int] = None,
        equity: bool = False,
    ):
        layer_types = [self.aggregation_dmg] + self.aggregated_metrics
        filtered_input_metrics = self._filter_metrics(metrics, type, rp, equity)
        return self._match_metrics_to_layers(filtered_input_metrics, layer_types)

    def _should_skip_metric(
        self, metric_name: str, rp: Optional[int], equity: bool
    ) -> bool:
        rp_match = re.search(r"RP(\d+)", metric_name)
        y_match = re.search(r"(\d+)Y", metric_name)
        name_match = rp_match or y_match

        if rp is None:
            if name_match:
                return True
            if not equity and "EW" in metric_name:
                return True
            if equity and "EW" not in metric_name:
                return True
        return False

    def _process_metric_name(
        self, metric_name: str, rp: Optional[int]
    ) -> tuple[str, bool]:
        rp_match = re.search(r"RP(\d+)", metric_name)
        y_match = re.search(r"(\d+)Y", metric_name)
        name_match = rp_match or y_match
        if rp is not None:
            if name_match:
                extracted_rp = int(name_match.group(1))
                name_check = extracted_rp == int(rp)
                cleaned_name = re.sub(r"(RP\d+|\d+Y)", "", metric_name)
                return cleaned_name.rstrip("_"), name_check
            else:
                return metric_name, False
        return metric_name, True

    def _filter_metrics(self, metrics, type, rp, equity):
        filtered = []
        for metric in metrics:
            metric_name = metric.get("name", "")
            metric["name_to_show"] = metric_name
            if type == "risk":
                if self._should_skip_metric(metric_name, rp, equity):
                    continue
                metric["name_to_show"], name_check = self._process_metric_name(
                    metric_name, rp
                )
                if rp is not None and not name_check:
                    continue
            filtered.append(metric)
        return filtered

    def _match_metrics_to_layers(self, metrics, layer_types):
        filtered_metrics = []
        for metric in metrics:
            for layer in layer_types:
                if layer.matches(metric):
                    filtered_metrics.append(
                        {
                            "metric": metric,
                            "bins": getattr(layer, "bins", None),
                            "colors": getattr(layer, "colors", None),
                            "decimals": getattr(layer, "decimals", None),
                        }
                    )
                    break
        return filtered_metrics


class VisualizationLayer(Layer):
    """The configuration of a layer to visualize in the gui.

    name : str
        The name of the layer to visualize.
    long_name : str
        The long name of the layer to visualize.
    path : str
        The path to the layer data to visualize.
    field_name : str
        The field names of the layer to visualize.
    decimals : Optional[int]
        The number of decimals to use for the layer to visualize. default is None.
    """

    name: str
    long_name: str
    path: str
    field_name: str
    decimals: Optional[int] = None


_DEFAULT_BIN_NR = 4


def interpolate_hex_colors(
    start_hex="#FFFFFF", end_hex="#860000", number_bins=_DEFAULT_BIN_NR
):
    """
    Interpolate between two hex colors and returns a list of number_bins hex color codes.

    Parameters
    ----------
        start_hex : str
            Starting color in hex format (e.g., "#FFFFFF").
        end_hex : str
            Ending color in hex format (e.g., "#000000").
        number_bins : int
            Number of colors to generate between the start and end colors.

    Returns
    -------
        list[str]
            List of hex color codes interpolated between the start and end colors.
    """

    def hex_to_rgb(hex_color):
        hex_color = hex_color.lstrip("#")
        return tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))

    def rgb_to_hex(rgb_color):
        return "#{:02X}{:02X}{:02X}".format(*rgb_color)

    start_rgb = hex_to_rgb(start_hex)
    end_rgb = hex_to_rgb(end_hex)

    interpolated_colors = []
    for i in range(number_bins):
        ratio = i / (number_bins - 1) if number_bins > 1 else 0
        interpolated_rgb = tuple(
            int(start + (end - start) * ratio) for start, end in zip(start_rgb, end_rgb)
        )
        interpolated_colors.append(rgb_to_hex(interpolated_rgb))

    return interpolated_colors


class VisualizationLayers(BaseModel):
    """The configuration of the layers you might want to visualize in the gui.

    Attributes
    ----------
    default : Layer
        The default layer settings the visualization layers.
    layers : list[VisualizationLayer]
        The layers to visualize.
    """

    buildings_min_zoom_level: int = 13
    layers: list[VisualizationLayer] = Field(default_factory=list)

    def add_layer(
        self,
        name: str,
        long_name: str,
        path: str,
        field_name: str,
        database_path: Path,
        decimals: Optional[int] = None,
        bins: Optional[list[float]] = None,
        colors: Optional[list[str]] = None,
    ) -> None:
        if not Path(path).is_absolute():
            raise ValueError(f"Path {path} must be absolute.")

        data = gpd.read_file(path)
        if field_name not in data.columns:
            raise ValueError(
                f"Field name {field_name} not found in data. Available fields: {data.columns.tolist()}"
            )

        if bins is None:
            _, _bins = pd.qcut(
                data[field_name], _DEFAULT_BIN_NR, retbins=True, duplicates="drop"
            )
            bins = _bins.tolist()[1:-1]

        if decimals is None:
            non_zero_bins = [abs(b) for b in bins if b != 0]
            min_non_zero = min(non_zero_bins) if non_zero_bins else 1
            decimals = max(int(-np.floor(np.log10(min_non_zero))), 0)

        if colors is None:
            nr_bins = len(bins) + 1
            colors = interpolate_hex_colors(number_bins=nr_bins)

        relative_path = Path(path).relative_to(database_path / "static")
        self.layers.append(
            VisualizationLayer(
                bins=bins,
                colors=colors,
                name=name,
                long_name=long_name,
                path=relative_path.as_posix(),
                field_name=field_name,
                decimals=decimals,
            )
        )


class GuiUnitModel(BaseModel):
    """The unit system used in the GUI.

    Attributes
    ----------
    default_length_units : us.UnitTypesLength
        The length units used in the GUI.
    default_distance_units : us.UnitTypesLength
        The distance units used in the GUI.
    default_area_units : us.UnitTypesArea
        The area units used in the GUI.
    default_volume_units : us.UnitTypesVolume
        The volume units used in the GUI.
    default_velocity_units : us.UnitTypesVelocity
        The velocity units used in the GUI.
    default_direction_units : us.UnitTypesDirection
        The direction units used in the GUI.
    default_discharge_units : us.UnitTypesDischarge
        The discharge units used in the GUI.
    default_intensity_units : us.UnitTypesIntensity
        The intensity units used in the GUI.
    default_cumulative_units : us.UnitTypesLength
        The cumulative units used in the GUI.
    """

    default_length_units: us.UnitTypesLength
    default_distance_units: us.UnitTypesLength
    default_area_units: us.UnitTypesArea
    default_volume_units: us.UnitTypesVolume
    default_velocity_units: us.UnitTypesVelocity
    default_direction_units: us.UnitTypesDirection
    default_discharge_units: us.UnitTypesDischarge
    default_intensity_units: us.UnitTypesIntensity
    default_cumulative_units: us.UnitTypesLength

    @staticmethod
    def imperial() -> "GuiUnitModel":
        return GuiUnitModel(
            default_length_units=us.UnitTypesLength.feet,
            default_distance_units=us.UnitTypesLength.miles,
            default_area_units=us.UnitTypesArea.sf,
            default_volume_units=us.UnitTypesVolume.cf,
            default_velocity_units=us.UnitTypesVelocity.mph,
            default_direction_units=us.UnitTypesDirection.degrees,
            default_discharge_units=us.UnitTypesDischarge.cfs,
            default_intensity_units=us.UnitTypesIntensity.inch_hr,
            default_cumulative_units=us.UnitTypesLength.inch,
        )

    @staticmethod
    def metric() -> "GuiUnitModel":
        return GuiUnitModel(
            default_length_units=us.UnitTypesLength.meters,
            default_distance_units=us.UnitTypesLength.meters,
            default_area_units=us.UnitTypesArea.m2,
            default_volume_units=us.UnitTypesVolume.m3,
            default_velocity_units=us.UnitTypesVelocity.mps,
            default_direction_units=us.UnitTypesDirection.degrees,
            default_discharge_units=us.UnitTypesDischarge.cms,
            default_intensity_units=us.UnitTypesIntensity.mm_hr,
            default_cumulative_units=us.UnitTypesLength.millimeters,
        )


class SyntheticTideModel(BaseModel):
    """Configuration for the synthetic tide.

    Parameters
    ----------
    harmonic_amplitude : us.UnitfulLength
        The amplitude of the tide harmonic relative to the datum.
    datum : str
        The datum to which the harmonic amplitude is relative.
        Should be a datum defined in `site.sfincs.waterlevels.datums`
    """

    harmonic_amplitude: us.UnitfulLength
    datum: str


class PlottingModel(BaseModel):
    """
    The configuration of the plotting in the gui.

    Attributes
    ----------
    excluded_datums : list[str]
        A list of datums that will be excluded from the forcing plot in event windows.
    synthetic_tide : SyntheticTideModel
        The configuration of the synthetic tide.
    """

    synthetic_tide: SyntheticTideModel
    excluded_datums: list[str] = Field(default_factory=list)


class GuiModel(BaseModel):
    """The accepted input for the variable gui in Site.

    Attributes
    ----------
    units : GuiUnitModel
        The unit system used in the GUI.
    output_layers : OutputLayers
        The configuration of the mapbox layers in the GUI.
    visualization_layers : VisualizationLayers
        The configuration of the visualization layers in the GUI.
    plotting : PlottingModel
        The configuration for creating hazard forcing plots.
    """

    units: GuiUnitModel
    output_layers: OutputLayers
    visualization_layers: VisualizationLayers
    plotting: PlottingModel

    @staticmethod
    def read_toml(path: Path) -> "GuiModel":
        with open(path, mode="rb") as fp:
            toml_contents = tomli.load(fp)

        return GuiModel(**toml_contents)



import os
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Type, TypeVar

import geopandas as gpd
import pyproj
import tomli
from pydantic import Field, field_serializer, field_validator, model_validator

from flood_adapt.config.site import Site
from flood_adapt.misc.utils import resolve_filepath, save_file_to_database
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.object_model import Object


class MeasureCategory(str, Enum):
    """Class describing the accepted input for the variable 'type' in Measure."""

    impact = "impact"
    hazard = "hazard"


class MeasureType(str, Enum):
    """Class describing the accepted input for the variable 'type' in Measure.

    Each type of measure is associated with a category (hazard or impact) and can be used to determine the type of measure.

    Attributes
    ----------
    floodwall : A floodwall measure.
    thin_dam : A thin dam measure.
    levee : A levee measure.
    pump : A pump measure.
    culvert : A culvert measure.
    water_square : A water square measure.
    greening : A greening measure.
    total_storage : A total storage measure.
    elevate_properties : An elevate properties measure.
    buyout_properties : A buyout properties measure.
    floodproof_properties : A floodproof properties measure.
    """

    # Hazard measures
    floodwall = "floodwall"
    thin_dam = "thin_dam"  # For now, same functionality as floodwall TODO: Add thin dam functionality
    levee = "levee"  # For now, same functionality as floodwall TODO: Add levee functionality
    pump = "pump"
    culvert = (
        "culvert"  # For now, same functionality as pump TODO: Add culvert functionality
    )
    water_square = "water_square"
    greening = "greening"
    total_storage = "total_storage"

    # Impact measures
    elevate_properties = "elevate_properties"
    buyout_properties = "buyout_properties"
    floodproof_properties = "floodproof_properties"

    @classmethod
    def is_hazard(cls, measure_type: str) -> bool:
        return measure_type in [
            cls.floodwall,
            cls.thin_dam,
            cls.levee,
            cls.pump,
            cls.culvert,
            cls.water_square,
            cls.greening,
            cls.total_storage,
        ]

    @classmethod
    def is_impact(cls, measure_type: str) -> bool:
        return measure_type in [
            cls.elevate_properties,
            cls.buyout_properties,
            cls.floodproof_properties,
        ]

    @classmethod
    def get_measure_category(cls, measure_type: str) -> MeasureCategory:
        if cls.is_hazard(measure_type):
            return MeasureCategory.hazard
        elif cls.is_impact(measure_type):
            return MeasureCategory.impact
        else:
            raise ValueError(f"Invalid measure type: {measure_type}")


class SelectionType(str, Enum):
    """Class describing the accepted input for the variable 'selection_type' in Measures.

    It is used to determine where to apply the measure to a model.

    Attributes
    ----------
    aggregation_area : Use aggregation area as geometry for the measure.
    polygon : Use polygon as geometry for the measure.
    polyline : Use polyline as geometry for the measure.
    all : Apply the measure to all geometries in the database.
    """

    aggregation_area = "aggregation_area"
    polygon = "polygon"
    polyline = "polyline"
    all = "all"


T = TypeVar("T", bound="Measure")


class Measure(Object):
    """The expected variables and data types of attributes common to all measures.

    A measure is a collection of attributes that can be applied to a model.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type: MeasureType
        Type of measure. Should be one of the MeasureType enum values.
    selection_type: SelectionType
        Type of selection. Should be one of the SelectionType enum values.
    polygon_file: str, Optional
        Path to a polygon file, either absolute or relative to the measure's toml path in the database.
    aggregation_area_name: str, Optional
        Name of the aggregation area. Required if `selection_type` is 'aggregation_area'.
    aggregation_area_type: str, Optional
        Type of aggregation area. Required if `selection_type` is 'aggregation_area'.
    """

    type: MeasureType
    selection_type: SelectionType

    polygon_file: Optional[str] = Field(
        default=None,
        min_length=1,
        description="Path to a polygon file, either absolute or relative to the measure path.",
    )

    aggregation_area_type: Optional[str] = None
    aggregation_area_name: Optional[str] = None

    @model_validator(mode="after")
    def validate_selection_type(self) -> "Measure":
        match self.selection_type:
            case SelectionType.all:
                pass
            case SelectionType.polygon | SelectionType.polyline:
                if not self.polygon_file:
                    raise ValueError(
                        "If `selection_type` is 'polygon' or 'polyline', then `polygon_file` needs to be set."
                    )
            case SelectionType.aggregation_area:
                if not self.aggregation_area_name:
                    raise ValueError(
                        "If `selection_type` is 'aggregation_area', then `aggregation_area_name` needs to be set."
                    )
                if not self.aggregation_area_type:
                    raise ValueError(
                        "If `selection_type` is 'aggregation_area', then `aggregation_area_type` needs to be set."
                    )
            case _:
                raise ValueError(
                    f"Invalid selection type: {self.selection_type}. "
                    "Must be one of 'aggregation_area', 'polygon', 'polyline', or 'all'."
                )
        return self

    @field_serializer("polygon_file")
    def serialize_polygon_file(self, value: Optional[str]) -> Optional[str]:
        """Serialize the polygon_file attribute to a string of only the file name."""
        if value is None:
            return None
        return Path(value).name

    @classmethod
    def load_file(cls: Type[T], file_path: Path | str | os.PathLike) -> T:
        """Load the measure from a file.

        Parameters
        ----------
        file_path : Path | str | os.PathLike
            Path to the file to load the measure from.

        Returns
        -------
        Measure
            The loaded measure object.
        """
        with open(file_path, mode="rb") as fp:
            toml = tomli.load(fp)
        measure = cls.model_validate(toml)

        if measure.polygon_file:
            measure.polygon_file = (
                Path(file_path).parent / measure.polygon_file
            ).as_posix()

        return measure

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        if self.polygon_file:
            Path(output_dir).mkdir(parents=True, exist_ok=True)
            src_path = resolve_filepath("measures", self.name, self.polygon_file)
            path = save_file_to_database(src_path, Path(output_dir))
            # Update the shapefile path in the object so it is saved in the toml file as well
            self.polygon_file = path.name


class HazardMeasure(Measure):
    """The expected variables and data types of attributes common to all hazard measures.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type: MeasureType
        Type of measure. Should be one of the MeasureType enum values and is_hazard.
    selection_type: SelectionType
        Type of selection. Should be one of the SelectionType enum values.
    polygon_file: str, Optional, default = None
        Path to a polygon file, either absolute or relative to the measure path in the database.

    """

    @field_validator("type")
    def validate_type(cls, value):
        if not MeasureType.is_hazard(value):
            raise ValueError(f"Invalid hazard type: {value}")
        return value


class ImpactMeasure(Measure):
    """The expected variables and data types of attributes common to all impact measures.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type: MeasureType
        Type of measure. Should be one of the MeasureType enum values and is_hazard.
    selection_type: SelectionType
        Type of selection. Should be one of the SelectionType enum values.
    polygon_file: str, Optional, default = None
        Path to a polygon file, either absolute or relative to the measure path in the database.
    property_type: str
        Type of property. Should be one of the PropertyType enum values.
    aggregation_area_type: str, Optional, default = None
        Type of aggregation area. Should be one of the SelectionType enum values.
    aggregation_area_name: str, Optional, default = None
            Name of the aggregation area.
    """

    property_type: str  # TODO make enum

    @field_validator("type")
    def validate_type(cls, value):
        if not MeasureType.is_impact(value):
            raise ValueError(f"Invalid impact type: {value}")
        return value


class Elevate(ImpactMeasure):
    """The expected variables and data types of the "elevate" impact measure.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type : MeasureType
        Type of measure. Should be "elevate_properties".
    selection_type : SelectionType
        Type of selection. Should be "polygon" or "aggregation_area".
    polygon_file : str, Optional
        Path to a polygon file, either absolute or relative to the measure path.
    aggregation_area_type : str, Optional
        Type of aggregation area. Should be "aggregation_area" or "all".
    aggregation_area_name : str, Optional
        Name of the aggregation area.
    property_type : str
        Type of property. Should be "residential" or "commercial".
    elevation : us.UnitfulLengthRefValue
        Elevation of the properties.
    """

    type: MeasureType = MeasureType.elevate_properties
    elevation: us.UnitfulLengthRefValue


class Buyout(ImpactMeasure):
    """The expected variables and data types of the "buyout" impact measure.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str, default ""
        Description of the measure.
    type : MeasureType, default MeasureType.buyout_properties
        Type of measure.
    selection_type : SelectionType
        Type of selection. Should be "polygon" or "aggregation_area".
    polygon_file : str, Optional
        Path to a polygon file, either absolute or relative to the measure path.
    aggregation_area_type : str, Optional
        Type of aggregation area. Should be "aggregation_area" or "all".
    aggregation_area_name : str, Optional
        Name of the aggregation area.
    property_type : str
        Type of property. Should be "residential" or "commercial".
    elevation : us.UnitfulLengthRefValue
        Elevation of the properties.

    """

    # Buyout has only the basic impact measure attributes
    type: MeasureType = MeasureType.buyout_properties


class FloodProof(ImpactMeasure):
    """The expected variables and data types of the "floodproof" impact measure.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type : MeasureType
        Type of measure. Should be "floodproof_properties".
    selection_type : SelectionType
        Type of selection. Should be "polygon" or "aggregation_area".
    polygon_file : str, Optional
        Path to a polygon file, either absolute or relative to the measure path.
    aggregation_area_type : str, Optional
        Type of aggregation area. Should be "aggregation_area" or "all".
    aggregation_area_name : str, Optional
        Name of the aggregation area.
    property_type : str
        Type of property. Should be "residential" or "commercial".
    elevation : us.UnitfulLengthRefValue
        Elevation of the properties.
    """

    type: MeasureType = MeasureType.floodproof_properties
    elevation: us.UnitfulLength


class FloodWall(HazardMeasure):
    """
    The expected variables and data types of the "floodwall" hazard measure.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type : MeasureType
        Type of measure. Should be "MeasureType.floodwall"
    selection_type : SelectionType
        Type of selection. Should be "SelectionType.polygon" or "SelectionType.aggregation_area".
    polygon_file : Optional[str]
        Path to a polygon file, either absolute or relative to the measure path.
    elevation : us.UnitfulLength
        Height of the floodwall.
    absolute_elevation : bool
        TODO remove?
    """

    type: MeasureType = MeasureType.floodwall
    elevation: us.UnitfulLength
    absolute_elevation: Optional[bool] = False


class Pump(HazardMeasure):
    """
    The expected variables and data types of the "pump" hazard measure.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type : MeasureType
        Type of measure. Should be "pump"
    selection_type : SelectionType
        Type of selection. Should be "polyline".
    polygon_file : str, Optional
        Path to a polygon file, either absolute or relative to the measure path.
    elevation : us.UnitfulLength
        Height of the floodwall.
    absolute_elevation : bool
        TODO remove?
    """

    type: MeasureType = MeasureType.pump
    discharge: us.UnitfulDischarge


class GreenInfrastructure(HazardMeasure):
    """The expected variables and data types of the "green infrastructure" hazard measure.

    Attributes
    ----------
    name: str
        Name of the measure.
    description: str
        Description of the measure.
    type : MeasureType
        Type of measure. Should be "greening"
    selection_type : SelectionType
        Type of selection. Should be "polygon" or "aggregation_area".
    height : us.UnitfulHeight, Optional
        Height of the green infrastructure.
    volume : us.UnitfulVolume, Optional
        Volume of the green infrastructure.
    polygon_file : str, Optional
        Path to a polygon file, either absolute or relative to the measure path.
    aggregation_area_type : str, Optional
        Type of aggregation area. Should be "aggregation_area".
    aggregation_area_name : str, Optional
        Name of the aggregation area.
    percent_area : float, Optional
        Percentage of the area that is green infrastructure.
    """

    type: MeasureType = MeasureType.greening
    volume: us.UnitfulVolume
    height: Optional[us.UnitfulHeight] = None
    aggregation_area_type: Optional[str] = None
    aggregation_area_name: Optional[str] = None
    percent_area: Optional[float] = Field(default=None, ge=0, le=100)

    @field_validator("height", mode="before", check_fields=False)
    def height_from_length(value: Any) -> Any:
        if isinstance(value, us.UnitfulLength):
            return us.UnitfulHeight(value=value.value, units=value.units)
        return value

    @model_validator(mode="after")
    def validate_hazard_type_values(self) -> "GreenInfrastructure":
        e_msg = f"Error parsing GreenInfrastructure: {self.name}"

        if self.type == MeasureType.total_storage:
            if self.height is not None or self.percent_area is not None:
                raise ValueError(
                    f"{e_msg}\nHeight and percent_area cannot be set for total storage type measures"
                )
            return self
        elif self.type == MeasureType.water_square:
            if self.percent_area is not None:
                raise ValueError(
                    f"{e_msg}\nPercentage_area cannot be set for water square type measures"
                )
            elif not isinstance(self.height, us.UnitfulHeight):
                raise ValueError(
                    f"{e_msg}\nHeight needs to be set for water square type measures"
                )
            return self
        elif self.type == MeasureType.greening:
            if not isinstance(self.height, us.UnitfulHeight) or not isinstance(
                self.percent_area, float
            ):
                raise ValueError(
                    f"{e_msg}\nHeight and percent_area needs to be set for greening type measures"
                )
        else:
            raise ValueError(
                f"{e_msg}\nType must be one of 'water_square', 'greening', or 'total_storage'"
            )
        return self

    @model_validator(mode="after")
    def validate_selection_type_values(self) -> "GreenInfrastructure":
        if self.selection_type == SelectionType.aggregation_area:
            if self.aggregation_area_name is None:
                raise ValueError(
                    "If `selection_type` is 'aggregation_area', then `aggregation_area_name` needs to be set."
                )
            if self.aggregation_area_type is None:
                raise ValueError(
                    "If `selection_type` is 'aggregation_area', then `aggregation_area_type` needs to be set."
                )
        return self

    @staticmethod
    def calculate_volume(
        area: us.UnitfulArea,
        height: us.UnitfulHeight,
        percent_area: float = 100.0,
    ) -> float:
        """Determine volume from area of the polygon and infiltration height.

        Parameters
        ----------
        area : us.UnitfulArea
            Area of polygon with units (calculated using calculate_polygon_area)
        height : us.UnitfulHeight
            Water height with units
        percent_area : float, optional
            Percentage area covered by green infrastructure [%], by default 100.0

        Returns
        -------
        float
            Volume [m3]
        """
        volume = (
            area.convert(us.UnitTypesArea.m2)
            * height.convert(us.UnitTypesLength.meters)
            * (percent_area / 100.0)
        )
        return volume

    @staticmethod
    def calculate_polygon_area(gdf: gpd.GeoDataFrame, site: Site) -> float:
        """Calculate area of a GeoDataFrame Polygon.

        Parameters
        ----------
        gdf : gpd.GeoDataFrame
            Polygon object
        site : Site
            site config (used for CRS)

        Returns
        -------
        area : float
            Area of the given polygon
        """
        # Determine local CRS
        crs = pyproj.CRS.from_string(site.sfincs.config.csname)
        gdf = gdf.to_crs(crs)

        # The GeoJSON file can contain multiple polygons
        polygon = gdf.geometry
        # Calculate the area of all polygons
        area = polygon.area.sum()
        return area



from pathlib import Path
from typing import Any, List, Type

import tomli

from flood_adapt.objects.forcing.discharge import (
    DischargeConstant,
    DischargeCSV,
    DischargeSynthetic,
)
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    ForcingType,
    IForcing,
    IForcingFactory,
)
from flood_adapt.objects.forcing.rainfall import (
    RainfallConstant,
    RainfallCSV,
    RainfallMeteo,
    RainfallNetCDF,
    RainfallSynthetic,
    RainfallTrack,
)
from flood_adapt.objects.forcing.waterlevels import (
    WaterlevelCSV,
    WaterlevelGauged,
    WaterlevelModel,
    WaterlevelSynthetic,
)
from flood_adapt.objects.forcing.wind import (
    WindConstant,
    WindCSV,
    WindMeteo,
    WindNetCDF,
    WindSynthetic,
    WindTrack,
)

__all__ = [
    "ForcingFactory",
    "ForcingSource",
    "ForcingType",
    "IForcing",
    "IForcingFactory",
    "WindConstant",
    "WindCSV",
    "WindMeteo",
    "WindSynthetic",
    "WindTrack",
    "WaterlevelCSV",
    "WaterlevelGauged",
    "WaterlevelModel",
    "WaterlevelSynthetic",
    "RainfallConstant",
    "RainfallCSV",
    "RainfallMeteo",
    "RainfallSynthetic",
    "RainfallTrack",
    "DischargeConstant",
    "DischargeCSV",
    "DischargeSynthetic",
]


class ForcingFactory(IForcingFactory):
    """Factory class for creating forcings."""

    WATERLEVELS: dict[ForcingSource, Type[IForcing]] = {
        ForcingSource.MODEL: WaterlevelModel,
        ForcingSource.CSV: WaterlevelCSV,
        ForcingSource.SYNTHETIC: WaterlevelSynthetic,
        ForcingSource.GAUGED: WaterlevelGauged,
    }

    WIND: dict[ForcingSource, Type[IForcing]] = {
        ForcingSource.METEO: WindMeteo,
        ForcingSource.TRACK: WindTrack,
        ForcingSource.CSV: WindCSV,
        ForcingSource.SYNTHETIC: WindSynthetic,
        ForcingSource.CONSTANT: WindConstant,
        ForcingSource.NETCDF: WindNetCDF,
    }

    RAINFALL: dict[ForcingSource, Type[IForcing]] = {
        ForcingSource.METEO: RainfallMeteo,
        ForcingSource.TRACK: RainfallTrack,
        ForcingSource.CSV: RainfallCSV,
        ForcingSource.SYNTHETIC: RainfallSynthetic,
        ForcingSource.CONSTANT: RainfallConstant,
        ForcingSource.NETCDF: RainfallNetCDF,
    }

    DISCHARGE: dict[ForcingSource, Type[IForcing]] = {
        ForcingSource.CSV: DischargeCSV,
        ForcingSource.SYNTHETIC: DischargeSynthetic,
        ForcingSource.CONSTANT: DischargeConstant,
    }

    FORCINGTYPES: dict[ForcingType, dict[ForcingSource, Type[IForcing]]] = {
        ForcingType.WATERLEVEL: WATERLEVELS,
        ForcingType.RAINFALL: RAINFALL,
        ForcingType.WIND: WIND,
        ForcingType.DISCHARGE: DISCHARGE,
    }

    @classmethod
    def read_forcing(
        cls,
        filepath: Path,
    ) -> tuple[Type[IForcing], ForcingType, ForcingSource]:
        """Extract forcing type and source from a TOML file."""
        with open(filepath, mode="rb") as fp:
            toml_data = tomli.load(fp)
        type = toml_data.get("type")
        source = toml_data.get("source")

        if type is None or source is None:
            raise ValueError(
                f"Forcing type {type} or source {source} not found in {filepath}"
            )
        forcing_cls = cls.get_forcing_class(ForcingType(type), ForcingSource(source))
        return forcing_cls, ForcingType(type), ForcingSource(source)

    @classmethod
    def get_forcing_class(
        cls, type: ForcingType, source: ForcingSource
    ) -> Type[IForcing]:
        """Get the forcing class corresponding to the type and source."""
        if (sources := cls.FORCINGTYPES.get(type)) is None:
            raise ValueError(f"Invalid forcing type: {type}")

        if (forcing_cls := sources.get(source)) is None:
            raise ValueError(
                f"Invalid forcing source: {source} for forcing type: {type}"
            )
        return forcing_cls

    @classmethod
    def load_file(cls, toml_file: Path) -> IForcing:
        """Create a forcing object from a TOML file."""
        with open(toml_file, mode="rb") as fp:
            toml_data = tomli.load(fp)
        return cls.load_dict(toml_data)

    @classmethod
    def load_dict(cls, attrs: dict[str, Any] | IForcing) -> IForcing:
        """Create a forcing object from a dictionary of attributes."""
        if isinstance(attrs, IForcing):
            return attrs
        type = attrs.get("type")
        source = attrs.get("source")
        if type is None or source is None:
            raise ValueError(
                f"Forcing type {type} or source {source} not found in attributes."
            )
        return cls.get_forcing_class(
            ForcingType(type), ForcingSource(source)
        ).model_validate(attrs)

    @classmethod
    def list_forcings(cls) -> List[Type[IForcing]]:
        """List all available forcing classes."""
        forcing_classes = set()
        for source_map in cls.FORCINGTYPES.values():
            for forcing in source_map.values():
                if forcing is not None:
                    forcing_classes.add(forcing)
        return list(forcing_classes)

    @classmethod
    def list_forcing_types_and_sources(cls) -> List[tuple[ForcingType, ForcingSource]]:
        """List all available forcing classes using a tuple of ForcingType and ForcingSource."""
        # TODO remove this when the backend supports all forcings
        ONLY_BACKEND_FORCINGS = {(ForcingType.WIND, ForcingSource.SYNTHETIC)}
        combinations = set()
        for type, source_map in cls.FORCINGTYPES.items():
            for source in source_map.keys():
                if (type, source) in ONLY_BACKEND_FORCINGS:
                    continue
                combinations.add((type, source))
        return list(combinations)



import logging
import math
import os
import shutil
import subprocess
from pathlib import Path
from typing import Any, Optional, Union

import geopandas as gpd
import pandas as pd
import tomli
from fiat_toolbox import FiatColumns, get_fiat_columns
from fiat_toolbox.equity.equity import Equity
from fiat_toolbox.infographics.infographics_factory import InforgraphicFactory
from fiat_toolbox.metrics_writer.fiat_write_metrics_file import MetricsFileWriter
from fiat_toolbox.metrics_writer.fiat_write_return_period_threshold import (
    ExceedanceProbabilityCalculator,
)
from fiat_toolbox.spatial_output.aggregation_areas import AggregationAreas
from fiat_toolbox.spatial_output.footprints import Footprints
from fiat_toolbox.utils import extract_variables, matches_pattern, replace_pattern
from hydromt_fiat.fiat import FiatModel

from flood_adapt.adapter.interface.impact_adapter import IImpactAdapter
from flood_adapt.config.fiat import FiatConfigModel
from flood_adapt.config.impacts import FloodmapType
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.path_builder import (
    ObjectDir,
)
from flood_adapt.misc.utils import cd, resolve_filepath
from flood_adapt.objects.events.events import Mode
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.measures.measures import (
    Buyout,
    Elevate,
    FloodProof,
    Measure,
    MeasureType,
)
from flood_adapt.objects.projections.projections import Projection
from flood_adapt.objects.scenarios.scenarios import Scenario

# Define naming structure for saved files
_IMPACT_COLUMNS = FiatColumns(
    object_id="Object ID",
    object_name="Object Name",
    primary_object_type="Primary Object Type",
    secondary_object_type="Secondary Object Type",
    extraction_method="Extraction Method",
    ground_floor_height="Ground Floor Height",
    ground_elevation="Ground Elevation",
    damage_function="Damage Function: {name}",
    max_potential_damage="Max Potential Damage: {name}",
    aggregation_label="Aggregation Label: {name}",
    inundation_depth="Inundation Depth",
    inundation_depth_rp="Inundation Depth ({years}Y)",
    reduction_factor="Reduction Factor",
    reduction_factor_rp="Reduction Factor ({years}Y)",
    damage="Damage: {name}",
    damage_rp="Damage: {name} ({years}Y)",
    total_damage="Total Damage",
    total_damage_rp="Total Damage ({years}Y)",
    risk_ead="Risk (EAD)",
    segment_length="Segment Length",
)

# Define column naming of FIAT model
_FIAT_COLUMNS: FiatColumns = get_fiat_columns(
    fiat_version="0.2.1"
)  # columns of FIAT # TODO add version from config

logger = FloodAdaptLogging.getLogger("FiatAdapter")


class FiatAdapter(IImpactAdapter):
    """
    ImpactAdapter for Delft-FIAT.

    It includes:
    - preprocessing methods for adding measures, projections and hazards
    - executing method for running a Delft-FIAT simulation
    - postprocessing methods for saving impact results
    """

    # TODO deal with all the relative paths for the files used
    # TODO IImpactAdapter and general Adapter class should NOT use the database

    _model: Optional[FiatModel] = None
    config: Optional[FiatConfigModel] = None
    exe_path: Optional[os.PathLike] = None
    fiat_columns: FiatColumns
    impact_columns: FiatColumns

    def __init__(
        self,
        model_root: Path,
        config: Optional[FiatConfigModel] = None,
        exe_path: Optional[os.PathLike] = None,
        delete_crashed_runs: bool = True,
        config_base_path: Optional[os.PathLike] = None,
    ) -> None:
        # TODO should exe_path and delete_crashed_runs be part of the config?
        # Load FIAT template
        self.config = config
        self.config_base_path = config_base_path
        self.exe_path = exe_path
        self.delete_crashed_runs = delete_crashed_runs
        self._model_root = model_root.resolve().as_posix()
        self.fiat_columns = _FIAT_COLUMNS
        self.impact_columns = _IMPACT_COLUMNS  # columns of FA impact output

    @property
    def model(self) -> FiatModel:
        """Lazily load and cache the FiatModel."""
        if self._model is None:
            self._model = FiatModel(root=self._model_root, mode="r")
            self._model.read()
        return self._model

    @property
    def model_root(self):
        return Path(self.model.root)

    @property
    def damage_types(self):
        """Get the damage types that are present in the exposure."""
        types = []
        for col in self.model.exposure.exposure_db.columns:
            if matches_pattern(col, self.fiat_columns.damage_function):
                name = extract_variables(col, self.fiat_columns.damage_function)["name"]
                types.append(name)
        return types

    def read(self, path: Path) -> None:
        """Read the fiat model from the current model root."""
        if Path(self.model.root).resolve() != Path(path).resolve():
            self.model.set_root(root=path.as_posix(), mode="r")
        self.model.read()

    def write(self, path_out: Union[str, os.PathLike], overwrite: bool = True) -> None:
        """Write the fiat model configuration to a directory."""
        if not isinstance(path_out, Path):
            path_out = Path(path_out).resolve()

        if not path_out.exists():
            path_out.mkdir(parents=True)

        write_mode = "w+" if overwrite else "w"
        with cd(path_out):
            self.model.set_root(root=path_out.as_posix(), mode=write_mode)
            self.model.write()

    def close_files(self):
        """Close all open files and clean up file handles."""
        loggers = [logger]
        if self._model is not None:
            loggers.append(self._model.logger)

        for _logger in loggers:
            if hasattr(_logger, "handlers"):
                for handler in _logger.handlers:
                    if isinstance(handler, logging.FileHandler):
                        handler.close()
                        _logger.removeHandler(handler)

    def __enter__(self) -> "FiatAdapter":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        self.close_files()
        return False

    def has_run(self, scenario: Scenario) -> bool:
        # TODO this should include a check for all output files , and then maybe save them as output paths and types
        """
        Check if the impact results file for the given scenario exists.

        Parameters
        ----------
        scenario : Scenario
            The scenario for which to check the FIAT results.

        Returns
        -------
        bool
            True if the FIAT results file exists, False otherwise.
        """
        impacts_path = self.database.get_impacts_path(scenario_name=scenario.name)
        fiat_results_path = impacts_path.joinpath(
            f"Impacts_detailed_{scenario.name}.csv"
        )
        return fiat_results_path.exists()

    def delete_model(self):
        """
        Delete the Delft-FIAT simulation folder.

        This method attempts to delete the directory specified by `self.model_root`.

        Raises
        ------
            OSError: If the directory cannot be deleted.
        """
        self.close_files()
        if self.model_root.exists():
            logger.info(f"Deleting {self.model_root}")
            shutil.rmtree(self.model_root, ignore_errors=True)

    def fiat_completed(self) -> bool:
        """Check if fiat has run as expected.

        Returns
        -------
        boolean
            True if fiat has run, False if something went wrong
        """
        log_file = self.model_root.joinpath(
            self.model.config["output"]["path"], "fiat.log"
        )
        if not log_file.exists():
            return False
        try:
            with open(log_file, "r", encoding="cp1252") as f:
                return "Geom calculation are done!" in f.read()
        except Exception as e:
            logger.error(f"Error while checking if FIAT has run: {e}")
            return False

    def preprocess(self, scenario: Scenario) -> None:
        """
        Preprocess the FIAT-model given a scenario by setting up projections, measures, and hazards, and then saves any changes made to disk.

        Args:
            scenario (Scenario): The scenario to preprocess, which includes projection,
                                  strategy, and hazard.

        Returns
        -------
            None
        """
        logger.info("Pre-processing Delft-FIAT model")
        # Projection
        projection = self.database.projections.get(scenario.projection)
        self.add_projection(projection)

        # Measures
        strategy = self.database.strategies.get(scenario.strategy)
        for measure in strategy.get_impact_measures():
            self.add_measure(measure)

        # Hazard
        floodmap = self.database.get_floodmap(scenario.name)
        var = "risk_maps" if floodmap.mode == Mode.risk else "zsmax"
        is_risk = floodmap.mode == Mode.risk
        self.set_hazard(
            map_fn=floodmap.paths,
            map_type=floodmap.map_type,
            var=var,
            is_risk=is_risk,
            units=us.UnitTypesLength.meters,
        )

        # Save any changes made to disk as well
        output_path = (
            self.database.get_impacts_path(scenario_name=scenario.name) / "fiat_model"
        )
        self.write(path_out=output_path)

    def run(self, scenario) -> None:
        """
        Execute the full process for a given scenario, including preprocessing, executing the simulation, and postprocessing steps.

        Args:
            scenario: An object containing the scenario data.

        Returns
        -------
            None
        """
        sim_path = (
            self.database.get_impacts_path(scenario_name=scenario.name) / "fiat_model"
        )

        self.preprocess(scenario)
        self.execute(sim_path)
        self.postprocess(scenario)

    def execute(
        self,
        path: Optional[os.PathLike] = None,
        exe_path: Optional[os.PathLike] = None,
        delete_crashed_runs: Optional[bool] = None,
        strict=True,
    ) -> bool:
        """
        Execute the FIAT model.

        Parameters
        ----------
        path : Optional[os.PathLike], optional
            The path to the model directory. If not provided, defaults to `self.model_root`.
        exe_path : Optional[os.PathLike], optional
            The path to the FIAT executable. If not provided, defaults to `self.exe_path`.
        delete_crashed_runs : Optional[bool], optional
            Whether to delete files from crashed runs. If not provided, defaults to `self.delete_crashed_runs`.
        strict : bool, optional
            Whether to raise an error if the FIAT model fails to run. Defaults to True.

        Returns
        -------
        bool
            True if the FIAT model run successfully, False otherwise.

        Raises
        ------
        ValueError
            If `exe_path` is not provided and `self.exe_path` is None.
        RuntimeError
            If the FIAT model fails to run and `strict` is True.
        """
        if path is None:
            path = self.model_root
        if exe_path is None:
            if self.exe_path is None:
                raise ValueError(
                    "'exe_path' needs to be provided either when calling FiatAdapter.execute() or during initialization of the FiatAdapter object."
                )
            exe_path = self.exe_path
        if delete_crashed_runs is None:
            delete_crashed_runs = self.delete_crashed_runs
        path = Path(path)
        fiat_log = path / "fiat.log"
        with cd(path):
            with FloodAdaptLogging.to_file(file_path=fiat_log):
                FiatAdapter._ensure_correct_hash_spacing_in_csv(path)

                logger.info(f"Running FIAT in {path}")
                process = subprocess.run(
                    args=[Path(exe_path).resolve().as_posix(), "run", "settings.toml"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                )
                logger.debug(process.stdout)

        if process.returncode != 0:
            if delete_crashed_runs:
                # Remove all files in the simulation folder except for the log files
                for subdir, dirs, files in os.walk(path, topdown=False):
                    for file in files:
                        if not file.endswith(".log"):
                            os.remove(os.path.join(subdir, file))

                    if not os.listdir(subdir):
                        shutil.rmtree(subdir, ignore_errors=True)

            if strict:
                raise RuntimeError(f"FIAT model failed to run in {path}.")
            else:
                logger.error(f"FIAT model failed to run in {path}.")

        if process.returncode == 0:
            self.read_outputs()

        return process.returncode == 0

    def read_outputs(self) -> None:
        """
        Read the output FIAT CSV file specified in the model configuration and stores the data in the `outputs` attribute.

        Attributes
        ----------
        outputs : dict
            A dictionary containing the following keys:
            - "path" : Path
                The path to the output directory.
            - "table" : DataFrame
                The contents of the output CSV file.
        """
        # Get output path
        outputs_path = self.model_root.joinpath(self.model.config["output"]["path"])

        # Get all csvs and concatenate them in a single table
        csv_outputs_df = []
        for output_csv in self.model.config["output"]["csv"]:
            csv_path = outputs_path.joinpath(
                self.model.config["output"]["csv"][output_csv]
            )
            output_csv_df = pd.read_csv(csv_path)
            csv_outputs_df.append(output_csv_df)
        output_csv = pd.concat(csv_outputs_df)

        # Store them
        self.outputs = {}
        self.outputs["path"] = outputs_path
        self.outputs["table"] = output_csv

    def _get_aggr_ind(self, aggr_label: str):
        """
        Retrieve the index of the aggregation configuration that matches the given label.

        Parameters
        ----------
        aggr_label : str
            The label of the aggregation to find.

        Returns
        -------
        int
            The index of the aggregation configuration that matches the given label.

        Raises
        ------
        IndexError
            If no aggregation with the given label is found.
        """
        ind = [
            i
            for i, aggr in enumerate(self.config.aggregation)
            if aggr.name == aggr_label
        ][0]

        return ind

    def postprocess(self, scenario):
        """
        Post-process the results of the Delft-FIAT simulation for a given scenario.

        Parameters
        ----------
        scenario : Scenario
            The scenario object containing all relevant data and configurations.

        Raises
        ------
        RuntimeError
            If the Delft-FIAT simulation did not run successfully.

        Post-processing steps include:
        - Reading the outputs of the Delft-FIAT simulation.
        - Adding exceedance probabilities for risk mode scenarios.
        - Saving detailed impacts per object to a CSV file.
        - Creating infometrics files based on different metric configurations.
        - Generating infographic files if configured.
        - Calculating equity-based damages for risk mode scenarios.
        - Saving aggregated metrics to shapefiles.
        - Merging points data to building footprints.
        - Creating a roads spatial file if configured.
        - Deleting the simulation folder if the site configuration is set to not keep the simulation.

        Logging
        -------
        Logs the start and completion of the post-processing steps.
        """
        if not self.fiat_completed():
            raise RuntimeError("Delft-FIAT did not run successfully!")

        logger.info("Post-processing Delft-FIAT results")

        if not self.outputs:
            self.read_outputs()
        mode = self.database.events.get(scenario.event).mode

        # Define scenario output path
        impacts_output_path = self.database.get_impacts_path(
            scenario_name=scenario.name
        )

        # Create column mapping to update column names
        name_translation = {}
        for col in self.outputs["table"].columns:  # iterate through output columns
            for field in list(self.impact_columns.model_fields):  # check for each field
                fiat_col = getattr(self.fiat_columns, field)
                if matches_pattern(col, fiat_col):
                    impact_col = getattr(self.impact_columns, field)
                    new_col = replace_pattern(col, fiat_col, impact_col)
                    if (
                        ".0Y" in new_col
                    ):  # TODO for now quick fix to account for float RP years, while metrics have integers
                        new_col = new_col.replace(".0Y", "Y")
                    name_translation[col] = new_col  # save mapping
        self.name_mapping = name_translation

        # Rename save outputs
        self.outputs["table"] = self.outputs["table"].rename(columns=self.name_mapping)

        # Save impacts per object
        fiat_results_path = impacts_output_path.joinpath(
            f"Impacts_detailed_{scenario.name}.csv"
        )
        self.outputs["table"].to_csv(fiat_results_path, index=False)

        # Add exceedance probabilities if needed (only for risk)
        if mode == Mode.risk:
            # Get config path
            # TODO check where this configs should be read from
            config_path = self.database.static_path.joinpath(
                "templates", "infometrics", "metrics_additional_risk_configs.toml"
            )
            with open(config_path, mode="rb") as fp:
                config = tomli.load(fp)["flood_exceedance"]
            self.add_exceedance_probability(
                column=config[
                    "column"
                ],  # TODO check how to the correct version of column
                threshold=config["threshold"],
                period=config["period"],
            )

        # Create the infometrics files
        if mode == Mode.risk:
            ext = "_risk"
        else:
            ext = ""

        # Get options for metric configurations
        metric_types = ["mandatory", "additional"]  # these are checked always

        if self.config.infographics:  # if infographics are created
            metric_types += ["infographic"]

        metric_config_paths = [
            self.database.static_path.joinpath(
                "templates", "infometrics", f"{name}_metrics_config{ext}.toml"
            )
            for name in metric_types
        ]

        # Specify the metrics output path
        metrics_outputs_path = impacts_output_path.parent.joinpath(
            f"Infometrics_{scenario.name}.csv"
        )
        self.create_infometrics(metric_config_paths, metrics_outputs_path)

        # Get paths of created aggregated infometrics
        aggr_metrics_paths = list(
            metrics_outputs_path.parent.glob(f"{metrics_outputs_path.stem}_*.csv")
        )

        # Create the infographic files
        if self.config.infographics:
            config_base_path = self.database.static_path.joinpath(
                "templates", "Infographics"
            )
            self.create_infographics(
                name=scenario.name,
                output_base_path=impacts_output_path.parent,
                config_base_path=config_base_path,
                metrics_path=metrics_outputs_path,
                mode=mode,
            )

        # Calculate equity based damages
        if mode == Mode.risk:
            for file in aggr_metrics_paths:
                # Load metrics
                aggr_label = file.stem.split(f"{metrics_outputs_path.stem}_")[-1]
                self.add_equity(aggr_label=aggr_label, metrics_path=file)

        # Save aggregated metrics to shapefiles
        for file in aggr_metrics_paths:
            aggr_label = file.stem.split(f"{metrics_outputs_path.stem}_")[-1]
            output_path = impacts_output_path.joinpath(
                f"Impacts_aggregated_{scenario.name}_{aggr_label}.gpkg"
            )
            self.save_aggregation_spatial(
                aggr_label=aggr_label, metrics_path=file, output_path=output_path
            )

        # Merge points data to building footprints
        self.save_building_footprints(
            output_path=impacts_output_path.joinpath(
                f"Impacts_building_footprints_{scenario.name}.gpkg"
            )
        )

        # Create a roads spatial file
        if self.config.roads_file_name:
            self.save_roads(
                output_path=impacts_output_path.joinpath(
                    f"Impacts_roads_{scenario.name}.gpkg"
                )
            )

        logger.info("Delft-FIAT post-processing complete!")

        # If site config is set to not keep FIAT simulation, delete folder
        if not self.config.save_simulation:
            self.delete_model()

    def add_measure(self, measure: Measure):
        """
        Add and apply a specific impact measure to the properties of the FIAT model.

        Parameters
        ----------
        measure : Measure
            The impact measure to be applied. It can be of type Elevate, FloodProof, or Buyout.

        Notes
        -----
        The method logs the application of the measure and calls the appropriate method based on the measure type:
            - Elevate: Calls elevate_properties(measure)
            - FloodProof: Calls floodproof_properties(measure)
            - Buyout: Calls buyout_properties(measure)

            If the measure type is unsupported, a warning is logged.
        """
        logger.info(f"Applying impact measure '{measure.name}'")
        if isinstance(measure, Elevate):
            self.elevate_properties(measure)
        elif isinstance(measure, FloodProof):
            self.floodproof_properties(measure)
        elif isinstance(measure, Buyout):
            self.buyout_properties(measure)
        else:
            logger.warning(
                f"Skipping unsupported measure type {measure.__class__.__name__}"
            )

    def add_projection(self, projection: Projection):
        """
        Add the socioeconomic changes part of a projection to the FIAT model.

        Parameters
        ----------
        projection : Projection
            The projection object containing socioeconomic changes to be applied.

        Notes
        -----
        - Economic growth is applied to all existing buildings if specified.
        - New population growth areas are added if specified, taking into account
        economic growth.
        - Population growth is applied to existing objects if specified.
        """
        logger.info(
            f"Applying socioeconomic changes from projection '{projection.name}'"
        )
        socio_economic_change = projection.socio_economic_change

        ids_all_buildings = self.get_all_building_ids()

        # Implement socioeconomic changes if needed
        # First apply economic growth to existing objects
        if not math.isclose(socio_economic_change.economic_growth, 0, abs_tol=1e-6):
            self.apply_economic_growth(
                economic_growth=socio_economic_change.economic_growth,
                ids=ids_all_buildings,  #
            )

        # Then the new population growth area is added if provided
        # In the new areas, the economic growth is taken into account!
        # Order matters since for the pop growth new, we only want the economic growth!
        if not math.isclose(
            socio_economic_change.population_growth_new, 0, abs_tol=1e-6
        ):
            # Get path of new development area geometry
            area_path = resolve_filepath(
                object_dir=ObjectDir.projection,
                obj_name=projection.name,
                path=socio_economic_change.new_development_shapefile,
            )

            # Get DEM location for assigning elevation to new areas
            dem = (
                self.database.static_path
                / "dem"
                / self.database.site.sfincs.dem.filename
            )
            # Call adapter method to add the new areas
            self.apply_population_growth_new(
                population_growth=socio_economic_change.population_growth_new,
                ground_floor_height=socio_economic_change.new_development_elevation.value,
                elevation_type=socio_economic_change.new_development_elevation.type,
                area_path=area_path,
                ground_elevation=dem,
            )

        # Then apply population growth to existing objects
        if not math.isclose(
            socio_economic_change.population_growth_existing, 0, abs_tol=1e-6
        ):
            self.apply_population_growth_existing(
                population_growth=socio_economic_change.population_growth_existing,
                ids=ids_all_buildings,
            )

    def set_hazard(
        self,
        map_fn: Union[os.PathLike, list[os.PathLike]],
        map_type: FloodmapType,
        var: str,
        is_risk: bool = False,
        units: str = us.UnitTypesLength.meters,
    ) -> None:
        """
        Set the hazard map and type for the FIAT model.

        Parameters
        ----------
        map_fn : str
            The filename of the hazard map.
        map_type : FloodmapType
            The type of the flood map.
        var : str
            The variable name in the hazard map.
        is_risk : bool, optional
            Flag indicating if the map is a risk output. Defaults to False.
        units : str, optional
            The units of the hazard map. Defaults to us.UnitTypesLength.meters.
        """
        logger.info(f"Setting hazard to the {map_type} map {map_fn}")
        # Add the floodmap data to a data catalog with the unit conversion
        wl_current_units = us.UnitfulLength(value=1.0, units=units)
        conversion_factor = wl_current_units.convert(self.model.exposure.unit)

        self.model.setup_hazard(
            map_fn=[Path(p).as_posix() for p in map_fn],
            map_type=map_type.value,
            rp=None,
            crs=None,  # change this in new version (maybe to str(floodmap.crs.split(':')[1]))
            nodata=-999,  # change this in new version
            var=var,
            chunks="auto",
            risk_output=is_risk,
            unit_conversion_factor=conversion_factor,
        )

    # PROJECTIONS

    def apply_economic_growth(
        self, economic_growth: float, ids: Optional[list] = None
    ) -> None:
        """
        Apply economic growth to the FIAT-Model by adjusting the maximum potential damage values in the exposure database.

        This method updates the max potential damage values in the exposure database by
        applying a given economic growth rate. It can optionally filter the updates to
        specific objects identified by their IDs.

        Parameters
        ----------
        economic_growth : float
            The economic growth rate to apply, expressed as a percentage.
        ids : Optional[list], default=None
            A list of object IDs to which the economic growth should be applied. If None, the growth is applied to all buildings.
        """
        logger.info(f"Applying economic growth of {economic_growth} %.")
        # Get columns that include max damage
        damage_cols = [
            c
            for c in self.model.exposure.exposure_db.columns
            if matches_pattern(c, self.fiat_columns.max_potential_damage)
        ]

        # Get objects that are buildings (using site info)
        buildings_rows = ~self.model.exposure.exposure_db[
            self.fiat_columns.primary_object_type
        ].isin(self.config.non_building_names)

        # If ids are given use that as an additional filter
        if ids:
            buildings_rows = buildings_rows & self.model.exposure.exposure_db[
                self.fiat_columns.object_id
            ].isin(ids)

        # Update columns using economic growth value
        updated_max_pot_damage = self.model.exposure.exposure_db.copy()
        updated_max_pot_damage.loc[buildings_rows, damage_cols] *= (
            1.0 + economic_growth / 100.0
        )

        # update fiat model
        self.model.exposure.update_max_potential_damage(
            updated_max_potential_damages=updated_max_pot_damage
        )

    def apply_population_growth_existing(
        self, population_growth: float, ids: Optional[list[str]] = None
    ) -> None:
        """
        Apply population growth to the FIAT-Model by adjusting the existing max potential damage values for buildings.

        This method updates the max potential damage values in the exposure database by
        applying a given population growth rate. It can optionally filter the updates to
        specific objects identified by their IDs.

        Parameters
        ----------
        population_growth : float
            The population growth rate as a percentage.
        ids : Optional[list[str]]
            A list of object IDs to filter the updates. If None, the updates are applied to all buildings.
        """
        logger.info(f"Applying population growth of {population_growth} %.")
        # Get columns that include max damage
        damage_cols = [
            c
            for c in self.model.exposure.exposure_db.columns
            if matches_pattern(c, self.fiat_columns.max_potential_damage)
        ]

        # Get objects that are buildings (using site info)
        buildings_rows = ~self.model.exposure.exposure_db[
            self.fiat_columns.primary_object_type
        ].isin(self.config.non_building_names)

        # If ids are given use that as an additional filter
        if ids:
            buildings_rows = buildings_rows & self.model.exposure.exposure_db[
                self.fiat_columns.object_id
            ].isin(ids)

        # Update columns using economic growth value
        updated_max_pot_damage = self.model.exposure.exposure_db.copy()
        updated_max_pot_damage.loc[buildings_rows, damage_cols] *= (
            1.0 + population_growth / 100.0
        )

        # update fiat model
        self.model.exposure.update_max_potential_damage(
            updated_max_potential_damages=updated_max_pot_damage
        )

    def apply_population_growth_new(
        self,
        population_growth: float,
        ground_floor_height: float,
        elevation_type: str,
        area_path: str,
        ground_elevation: Union[None, str, Path] = None,
    ) -> None:
        """
        Apply population growth in a new area by adding new objects in the model.

        Parameters
        ----------
        population_growth : float
            The percentage of population growth to apply.
        ground_floor_height : float
            The height of the ground floor.
        elevation_type : str
            The type of elevation reference to use. Must be either 'floodmap' or 'datum'.
        area_path : str
            The path to the area file.
        ground_elevation : Union[None, str, Path], optional
            The ground elevation reference. Default is None.

        Raises
        ------
        ValueError
            If `elevation_type` is 'floodmap' and base flood elevation (bfe) map is not provided.
            If `elevation_type` is not 'floodmap' or 'datum'.
        """
        logger.info(
            f"Applying population growth of {population_growth} %, by creating a new development area using the geometries from {area_path} and a ground floor height of {ground_floor_height} {self.model.exposure.unit} above '{elevation_type}'."
        )
        # Get reference type to align with hydromt
        if elevation_type == "floodmap":
            if not self.config.bfe:
                raise ValueError(
                    "Base flood elevation (bfe) map is required to use 'floodmap' as reference."
                )
            kwargs = {
                "elevation_reference": "geom",
                "path_ref": self.database.static_path.joinpath(self.config.bfe.geom),
                "attr_ref": self.config.bfe.field_name,
            }
        elif elevation_type == "datum":
            kwargs = {"elevation_reference": "datum"}
        else:
            raise ValueError("elevation type can only be one of 'floodmap' or 'datum'")
        # Get aggregation areas info
        aggregation_areas = [
            self.database.static_path.joinpath(aggr.file)
            for aggr in self.config.aggregation
        ]
        attribute_names = [aggr.field_name for aggr in self.config.aggregation]
        label_names = [
            self.fiat_columns.aggregation_label.format(name=aggr.name)
            for aggr in self.config.aggregation
        ]
        new_dev_geom_name = Path(self.config.new_development_file_name).stem
        # Ensure new_devs geom is in the correct CRS
        new_dev_geom = gpd.read_file(area_path)
        if new_dev_geom.crs != self.model.exposure.crs:
            logger.warning(
                f"New development area geometries are in {new_dev_geom.crs}, but the model is in {self.model.exposure.crs}. Reprojecting geometries."
            )
            new_dev_geom = new_dev_geom.to_crs(self.model.exposure.crs)
        # Replace file with the reprojected one
        os.remove(area_path)
        new_dev_geom.to_file(area_path)
        # Use hydromt function
        self.model.exposure.setup_new_composite_areas(
            percent_growth=population_growth,
            geom_file=Path(area_path),
            ground_floor_height=ground_floor_height,
            damage_types=self.damage_types,
            vulnerability=self.model.vulnerability,
            ground_elevation=ground_elevation,
            aggregation_area_fn=aggregation_areas,
            attribute_names=attribute_names,
            label_names=label_names,
            geom_name=new_dev_geom_name,
            **kwargs,
        )

    # MEASURES
    @staticmethod
    def _get_area_name(measure: Measure):
        """
        Determine the area name based on the selection type of the measure.

        Parameters
        ----------
        measure : Measure
            An instance of Measure containing attributes that define the selection type and area.

        Returns
        -------
        str
            The name of the area. It returns the aggregation area name if the selection type is "aggregation_area",
            the polygon file name if the selection type is "polygon", and "all" for any other selection type.
        """
        if measure.selection_type == "aggregation_area":
            area = measure.aggregation_area_name
        elif measure.selection_type == "polygon":
            area = measure.polygon_file
        else:
            area = "all"
        return area

    def elevate_properties(self, elevate: Elevate) -> None:
        """
        Elevate the ground floor height of properties based on the provided Elevate measure.

        Parameters
        ----------
        elevate : Elevate
            The Elevate measure containing the elevation details.

        Raises
        ------
        ValueError
            If the elevation type is 'floodmap' and the base flood elevation (bfe) map is not provided.
            If the elevation type is not 'floodmap' or 'datum'.
        """
        area = self._get_area_name(elevate)
        logger.info(
            f"Elevating '{elevate.property_type}' type properties in '{area}' by {elevate.elevation} relative to '{elevate.elevation.type}'."
        )
        # If ids are given use that as an additional filter
        objectids = self.get_object_ids(elevate)

        # Get reference type to align with hydromt
        if elevate.elevation.type == "floodmap":
            if not self.config.bfe:
                raise ValueError(
                    "Base flood elevation (bfe) map is required to use 'floodmap' as reference."
                )
            if self.config.bfe.table:
                path_ref = self.config_base_path.joinpath(self.config.bfe.table)
                height_reference = "table"
            else:
                path_ref = self.config_base_path.joinpath(self.config.bfe.geom)
                height_reference = "geom"
            # Use hydromt function
            self.model.exposure.raise_ground_floor_height(
                raise_by=elevate.elevation.value,
                objectids=objectids,
                height_reference=height_reference,
                path_ref=path_ref,
                attr_ref=self.config.bfe.field_name,
            )

        elif elevate.elevation.type == "datum":
            # Use hydromt function
            self.model.exposure.raise_ground_floor_height(
                raise_by=elevate.elevation.value,
                objectids=objectids,
                height_reference="datum",
            )
        else:
            raise ValueError("elevation type can only be one of 'floodmap' or 'datum'")

    def buyout_properties(self, buyout: Buyout) -> None:
        """
        Apply the buyout measure to the properties by setting their maximum potential damage to zero.

        Parameters
        ----------
        buyout : Buyout
            The Buyout measure containing the details of the properties to be bought out.

        """
        area = self._get_area_name(buyout)
        logger.info(f"Buying-out '{buyout.property_type}' type properties in '{area}'.")
        # Get columns that include max damage
        damage_cols = [
            c
            for c in self.model.exposure.exposure_db.columns
            if matches_pattern(c, self.fiat_columns.max_potential_damage)
        ]

        # Get objects that are buildings (using site info)
        buildings_rows = ~self.model.exposure.exposure_db[
            self.fiat_columns.primary_object_type
        ].isin(self.config.non_building_names)

        # Get rows that are affected
        objectids = self.get_object_ids(buyout)
        rows = (
            self.model.exposure.exposure_db[self.fiat_columns.object_id].isin(objectids)
            & buildings_rows
        )

        # Update columns
        updated_max_pot_damage = self.model.exposure.exposure_db.copy()
        updated_max_pot_damage.loc[rows, damage_cols] *= 0

        # update fiat model
        self.model.exposure.update_max_potential_damage(
            updated_max_potential_damages=updated_max_pot_damage
        )

    def floodproof_properties(self, floodproof: FloodProof) -> None:
        """
        Apply floodproofing measures to the properties by truncating the damage function.

        Parameters
        ----------
        floodproof : FloodProof
            The FloodProof measure containing the details of the properties to be floodproofed.
        """
        area = self._get_area_name(floodproof)
        logger.info(
            f"Flood-proofing '{floodproof.property_type}' type properties in '{area}' by {floodproof.elevation}."
        )
        # If ids are given use that as an additional filter
        objectids = self.get_object_ids(floodproof)

        # Use hydromt function
        self.model.exposure.truncate_damage_function(
            objectids=objectids,
            floodproof_to=floodproof.elevation.value,
            damage_function_types=self.damage_types,
            vulnerability=self.model.vulnerability,
        )

    # STATIC METHODS

    def get_buildings(self) -> gpd.GeoDataFrame:
        """
        Retrieve the building geometries from the FIAT model's exposure database.

        Returns
        -------
        gpd.GeoDataFrame
            A GeoDataFrame containing the geometries of all buildings in the FIAT model.

        Raises
        ------
        ValueError
            If the FIAT model does not have an exposure database initialized.
        """
        if self.model.exposure is None:
            raise ValueError(
                "FIAT model does not have exposure, make sure your model has been initialized."
            )
        gdf_0 = self.model.exposure.select_objects(
            primary_object_type="ALL",
            non_building_names=self.config.non_building_names,
            return_gdf=True,
        )
        # Rename columns
        name_translation = {}
        for col in gdf_0.columns:  # iterate through output columns
            for field in list(self.impact_columns.model_fields):  # check for each field
                fiat_col = getattr(self.fiat_columns, field)
                if matches_pattern(col, fiat_col):
                    impact_col = getattr(self.impact_columns, field)
                    new_col = replace_pattern(col, fiat_col, impact_col)
                    name_translation[col] = new_col  # save mapping
        gdf = gdf_0.rename(columns=name_translation)
        return gdf

    def get_property_types(self) -> list:
        """
        Retrieve the list of property types from the FIAT model's exposure database.

        Returns
        -------
        list
            A list of property types available in the FIAT model.

        Raises
        ------
        ValueError
            If no property types are found in the FIAT model.
        """
        types = self.model.exposure.get_primary_object_type()
        if types is None:
            raise ValueError("No property types found in the FIAT model.")
        types.append("all")  # Add "all" type for using as identifier

        names = self.config.non_building_names
        if names:
            for name in names:
                if name in types:
                    types.remove(name)

        return types

    def get_all_building_ids(self):
        """
        Retrieve the IDs of all buildings in the FIAT model.

        Returns
        -------
        list
            A list of IDs for all buildings in the FIAT model.
        """
        # Get ids of existing buildings
        ids = self.model.exposure.get_object_ids(
            "all", non_building_names=self.config.non_building_names
        )
        return ids

    def get_object_ids(self, measure: Measure) -> list[Any]:
        """
        Retrieve the object IDs for a given impact measure.

        Parameters
        ----------
        measure : Measure
            The impact measure for which to retrieve object IDs.

        Returns
        -------
        list[Any]
            A list of object IDs that match the criteria of the given measure.

        Raises
        ------
        ValueError
            If the measure type is not an impact measure.
        """
        if not MeasureType.is_impact(measure.type):
            raise ValueError(
                f"Measure type {measure.type} is not an impact measure. "
                "Can only retrieve object ids for impact measures."
            )

        # check if polygon file is used, then get the absolute path
        if measure.polygon_file:
            polygon_file = resolve_filepath(
                object_dir=ObjectDir.measure,
                obj_name=measure.name,
                path=measure.polygon_file,
            ).as_posix()
        else:
            polygon_file = None

        # use the hydromt-fiat method to the ids
        ids = self.model.exposure.get_object_ids(
            selection_type=measure.selection_type,
            property_type=measure.property_type,
            non_building_names=self.config.non_building_names,
            aggregation=measure.aggregation_area_type,
            aggregation_area_name=measure.aggregation_area_name,
            polygon_file=polygon_file,
        )

        return ids

    # POST-PROCESSING METHODS

    def add_exceedance_probability(
        self, column: str, threshold: float, period: int
    ) -> pd.DataFrame:
        """Calculate exceedance probabilities and append them to the results table.

        Parameters
        ----------
        column : str
            The name of the column to calculate exceedance probabilities for.
        threshold : float
            The threshold value for exceedance probability calculation.
        period : int
            The return period for exceedance probability calculation.

        Returns
        -------
        pd.DataFrame
            The updated results table with exceedance probabilities appended.
        """
        logger.info("Calculating exceedance probabilities")
        fiat_results_df = ExceedanceProbabilityCalculator(column).append_probability(
            self.outputs["table"], threshold, period
        )
        self.outputs["table"] = fiat_results_df
        return self.outputs["table"]

    def create_infometrics(
        self, metric_config_paths: list[os.PathLike], metrics_output_path: os.PathLike
    ) -> None:
        """
        Create infometrics files based on the provided metric configuration paths.

        Parameters
        ----------
        metric_config_paths : list[os.PathLike]
            A list of paths to the metric configuration files.
        metrics_output_path : os.PathLike
            The path where the metrics output file will be saved.

        Raises
        ------
        FileNotFoundError
            If a mandatory metric configuration file does not exist.
        """
        # Get the metrics configuration
        logger.info("Calculating infometrics")

        # Write the metrics to file
        # Check if type of metric configuration is available
        for metric_file in metric_config_paths:
            if metric_file.exists():
                metrics_writer = MetricsFileWriter(
                    metric_file,
                    aggregation_label_fmt=self.impact_columns.aggregation_label,
                )

                metrics_writer.parse_metrics_to_file(
                    df_results=self.outputs["table"],
                    metrics_path=metrics_output_path,
                    write_aggregate=None,
                )

                metrics_writer.parse_metrics_to_file(
                    df_results=self.outputs["table"],
                    metrics_path=metrics_output_path,
                    write_aggregate="all",
                )
            else:
                if "mandatory" in metric_file.name.lower():
                    raise FileNotFoundError(
                        f"Mandatory metric configuration file {metric_file} does not exist!"
                    )

    def create_infographics(
        self,
        name: str,
        output_base_path: os.PathLike,
        config_base_path: os.PathLike,
        metrics_path: os.PathLike,
        mode: Mode = Mode.single_event,
    ):
        """Create infographic files based on the provided metrics and configuration.

        Parameters
        ----------
        name : str
            The name of the scenario.
        output_base_path : os.PathLike
            The base path where the output files will be saved.
        config_base_path : os.PathLike
            The base path where the configuration files are located.
        metrics_path : os.PathLike
            The path to the metrics file.
        mode : Mode, optional
            The mode of the infographic, by default Mode.single_event.
        """
        logger.info("Creating infographics")

        # Check if infographics config file exists
        if mode == Mode.risk:
            config_path = config_base_path.joinpath("config_risk_charts.toml")
            if not config_path.exists():
                logger.warning(
                    "Risk infographic cannot be created, since 'config_risk_charts.toml' is not available"
                )
                return

        # Get the infographic
        InforgraphicFactory.create_infographic_file_writer(
            infographic_mode=mode,
            scenario_name=name,
            metrics_full_path=metrics_path,
            config_base_path=config_base_path,
            output_base_path=output_base_path,
        ).write_infographics_to_file()

    def add_equity(
        self,
        aggr_label: str,
        metrics_path: os.PathLike,
        damage_column_pattern: str = "TotalDamageRP{rp}",
        gamma: float = 1.2,
    ):
        """Calculate equity-based damages for a given aggregation label.

        Parameters
        ----------
        aggr_label : str
            The label of the aggregation area.
        metrics_path : os.PathLike
            The path to the metrics file.
        damage_column_pattern : str, optional
            The pattern for the damage column names, by default "TotalDamageRP{rp}".
        gamma : float, optional
            The equity weight parameter, by default 1.2
        """
        # TODO gamma in configuration file?

        ind = self._get_aggr_ind(aggr_label)
        # TODO check what happens if aggr_label not in config

        if self.config.aggregation[ind].equity is None:
            logger.warning(
                f"Cannot calculate equity weighted risk for aggregation label: {aggr_label}, because equity inputs are not available."
            )
            return

        logger.info(
            f"Calculating equity weighted risk for aggregation label: {aggr_label} "
        )
        metrics = pd.read_csv(metrics_path)
        # Create Equity object
        equity = Equity(
            census_table=self.config_base_path.joinpath(
                self.config.aggregation[ind].equity.census_data
            ),
            damages_table=metrics,
            aggregation_label=self.config.aggregation[ind].field_name,
            percapitaincome_label=self.config.aggregation[
                ind
            ].equity.percapitaincome_label,
            totalpopulation_label=self.config.aggregation[
                ind
            ].equity.totalpopulation_label,
            damage_column_pattern=damage_column_pattern,
        )
        # Calculate equity
        df_equity = equity.equity_calculation(gamma)
        # Merge with metrics tables and resave
        metrics_new = metrics.merge(
            df_equity,
            left_on=metrics.columns[0],
            right_on=self.config.aggregation[ind].field_name,
            how="left",
        )
        del metrics_new[self.config.aggregation[ind].field_name]
        metrics_new = metrics_new.set_index(metrics_new.columns[0])
        metrics_new.loc["Description", ["EW", "EWEAD", "EWCEAD"]] = [
            "Equity weight",
            "Equity weighted  expected annual damage",
            "Equity weighted certainty equivalent  annual damage",
        ]
        metrics_new.loc["Show In Metrics Table", ["EW", "EWEAD", "EWCEAD"]] = [
            True,
            True,
            True,
        ]
        metrics_new.loc["Show In Metrics Map", ["EW", "EWEAD", "EWCEAD"]] = [
            False,
            True,
            False,
        ]
        metrics_new.loc["Long Name", ["EW", "EWEAD", "EWCEAD"]] = [
            "Equity weight",
            "Equity weighted  expected annual damage",
            "Equity weighted certainty equivalent  annual damage",
        ]
        metrics_new.index.name = None
        metrics_new.to_csv(metrics_path)

    def save_aggregation_spatial(
        self, aggr_label: str, metrics_path: os.PathLike, output_path: os.PathLike
    ):
        """
        Save aggregated metrics to a spatial file.

        Parameters
        ----------
        aggr_label : str
            The label of the aggregation area.
        metrics_path : os.PathLike
            The path to the metrics file.
        output_path : os.PathLike
            The path where the output spatial file will be saved.
        """
        logger.info(f"Saving impacts for aggregation areas type: '{aggr_label}'")

        metrics = pd.read_csv(metrics_path)

        # Load aggregation areas
        ind = self._get_aggr_ind(aggr_label)

        aggr_areas_path = self.config_base_path.joinpath(
            self.config.aggregation[ind].file
        )

        aggr_areas = gpd.read_file(aggr_areas_path, engine="pyogrio")

        # Save file
        AggregationAreas.write_spatial_file(
            metrics,
            aggr_areas,
            output_path,
            id_name=self.config.aggregation[ind].field_name,
            file_format="geopackage",
        )

    def save_building_footprints(self, output_path: os.PathLike):
        """
        Aggregate impacts at a building footprint level and then saves to an output file.

        Parameters
        ----------
        output_path : os.PathLike
            The path where the output spatial file will be saved.

        Raises
        ------
        ValueError
            If no building footprints are provided in the configuration.
        """
        logger.info("Calculating impacts at a building footprint scale")

        # Read the existing building points
        buildings = self.model.exposure.select_objects(
            primary_object_type="ALL",
            non_building_names=self.config.non_building_names,
            return_gdf=True,
        )

        # Change names
        buildings = buildings[[self.fiat_columns.object_id, "geometry"]]
        buildings = buildings.rename(
            columns={self.fiat_columns.object_id: self.impact_columns.object_id}
        )

        # Get all results per building
        fiat_results_df = gpd.GeoDataFrame(
            self.outputs["table"].merge(
                buildings,
                on=self.impact_columns.object_id,
                how="inner",
            )
        )

        # Check which footprint case we have
        # If FIAT has points and external footprints are provided
        if self.config.building_footprints:
            method = "external_footprints"
            # Get footprints file
            footprints_path = self.config_base_path.joinpath(
                self.config.building_footprints
            )
            # Read building footprints
            footprints_gdf = gpd.read_file(footprints_path, engine="pyogrio")
            field_name = "BF_FID"
        # If FIAT has footprints already
        elif all(buildings.geometry.geom_type.isin(["Polygon", "MultiPolygon"])):
            method = "internal_footprints"
            footprints_gdf = buildings[[self.impact_columns.object_id, "geometry"]]
            field_name = self.impact_columns.object_id
        # If FIAT has points and no external footprints are available
        else:
            method = "no_footprints"

        # Based on case follow different workflow
        if method in ["external_footprints", "internal_footprints"]:
            footprints = Footprints(
                footprints=footprints_gdf,
                fiat_columns=self.impact_columns,
                field_name=field_name,
            )
            footprints.aggregate(fiat_results_df)
        elif method == "no_footprints":
            footprints = Footprints(fiat_columns=self.impact_columns)
            footprints.set_point_data(fiat_results_df)

        # Normalize damages
        footprints.calc_normalized_damages()

        # Save footprint
        footprints.write(output_path)

    def save_roads(self, output_path: os.PathLike):
        """
        Save the impacts on roads to a spatial file.

        Parameters
        ----------
        output_path : os.PathLike
            The path where the output spatial file will be saved.
        """
        logger.info("Calculating road impacts")
        # Read roads spatial file
        roads = gpd.read_file(
            self.outputs["path"].joinpath(self.config.roads_file_name)
        )
        roads = roads.rename(columns=self.name_mapping)
        # Get columns to use
        aggr_cols = [
            name
            for name in self.outputs["table"].columns
            if self.impact_columns.aggregation_label in name
        ]
        inun_cols = [
            name
            for name in roads.columns
            if self.impact_columns.inundation_depth in name
        ]
        # Merge data
        roads = roads[[self.impact_columns.object_id, "geometry"] + inun_cols].merge(
            self.outputs["table"][
                [self.impact_columns.object_id, self.impact_columns.primary_object_type]
                + aggr_cols
            ],
            on=self.impact_columns.object_id,
        )
        # Save as geopackage
        roads.to_file(output_path, driver="GPKG")

    @staticmethod
    def _ensure_correct_hash_spacing_in_csv(
        model_root: Path, hash_spacing: int = 1
    ) -> None:
        """
        Ensure that the CSV file has the correct number of spaces between hashes.

        When writing csv files, FIAT does not add spaces between the hashes and the line, which leads to errors on linux.


        Parameters
        ----------
        file_path : Path
            The path to the model root.
        hash_spacing : int, optional
            The number of spaces between hashes, by default 1.
        """
        for dirpath, _, filenames in os.walk(model_root):
            for filename in filenames:
                if not filename.lower().endswith(".csv"):
                    continue
                file_path = os.path.join(dirpath, filename)

                with open(file_path, "r") as file:
                    lines = file.readlines()

                with open(file_path, "w") as file:
                    for line in lines:
                        if line.startswith("#"):
                            line = "#" + " " * hash_spacing + line.lstrip("#")
                        file.write(line)

    def _delete_simulation_folder(self, scn: Scenario):
        """
        Delete the Delft-FIAT simulation folder for a given scenario.

        Parameters
        ----------
        scn : Scenario
            The scenario for which the simulation folder should be deleted.

        Raises
        ------
        OSError
            If the directory cannot be deleted.
        """
        simulation_path = (
            self.database.scenarios.output_path / scn.name / "Impacts" / "fiat_model"
        )
        if simulation_path.exists():
            self.close_files()
            shutil.rmtree(simulation_path, ignore_errors=True)
            logger.info(f"Deleted Delft-FIAT simulation folder: {simulation_path}")



from typing import ClassVar, List

from flood_adapt.objects.events.events import Event, Template
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    ForcingType,
)
from flood_adapt.objects.forcing.time_frame import TimeFrame

__all__ = ["HistoricalEvent", "TimeFrame"]


class HistoricalEvent(Event):
    """BaseModel describing the expected variables and data types for parameters of HistoricalEvent that extend the parent class Event.

    Attributes
    ----------
    name : str
        The name of the event.
    description : str
        The description of the event. Defaults to "".
    time : TimeFrame
        The time frame of the event.
    template : Template
        The template of the event. Defaults to Template.Historical.
    mode : Mode
        The mode of the event. Defaults to Mode.single_event.
    rainfall_multiplier : float
        The rainfall multiplier of the event. Defaults to 1.0.
    forcings : dict[ForcingType, list[IForcing]]
        The forcings of the event.
    """

    ALLOWED_FORCINGS: ClassVar[dict[ForcingType, List[ForcingSource]]] = {
        ForcingType.RAINFALL: [
            ForcingSource.CSV,
            # ForcingSource.METEO, # Temporarily excluded due to bug in hydromt-sfincs. fixed in v1.3.0
            ForcingSource.SYNTHETIC,
            ForcingSource.CONSTANT,
        ],
        ForcingType.WIND: [
            ForcingSource.CSV,
            # ForcingSource.METEO, # Temporarily excluded due to bug in hydromt-sfincs. fixed in v1.3.0
            ForcingSource.CONSTANT,
        ],
        ForcingType.WATERLEVEL: [
            # ForcingSource.MODEL, # Temporarily excluded due to the METEO bug in hydromt-sfincs. fixed in v1.3.0
            ForcingSource.CSV,
            ForcingSource.SYNTHETIC,
            ForcingSource.GAUGED,
        ],
        ForcingType.DISCHARGE: [
            ForcingSource.CSV,
            ForcingSource.SYNTHETIC,
            ForcingSource.CONSTANT,
        ],
    }
    template: Template = Template.Historical



from pathlib import Path
from tempfile import gettempdir
from typing import List, Optional

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from flood_adapt.config.site import Site
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.path_builder import (
    db_path,
)
from flood_adapt.objects.events.events import Event, Template
from flood_adapt.objects.forcing.discharge import (
    DischargeConstant,
    DischargeCSV,
    DischargeSynthetic,
)
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    ForcingType,
    IDischarge,
)
from flood_adapt.objects.forcing.rainfall import (
    RainfallConstant,
    RainfallCSV,
    RainfallSynthetic,
)
from flood_adapt.objects.forcing.waterlevels import (
    WaterlevelCSV,
    WaterlevelGauged,
    WaterlevelSynthetic,
)
from flood_adapt.objects.forcing.wind import (
    WindConstant,
    WindCSV,
    WindSynthetic,
)

# TODO remove from frontend
UNPLOTTABLE_SOURCES = [ForcingSource.TRACK, ForcingSource.METEO, ForcingSource.MODEL]
logger = FloodAdaptLogging.getLogger("Plotting")


def plot_forcing(
    event: Event,
    site: Site,
    forcing_type: ForcingType,
) -> tuple[str, Optional[List[Exception]]]:
    """Plot the forcing data for the event."""
    if event.forcings.get(forcing_type) is None:
        return "", None

    match forcing_type:
        case ForcingType.RAINFALL:
            return plot_rainfall(event, site)
        case ForcingType.WIND:
            return plot_wind(event, site)
        case ForcingType.WATERLEVEL:
            return plot_waterlevel(event, site)
        case ForcingType.DISCHARGE:
            return plot_discharge(event, site)
        case _:
            raise NotImplementedError(
                "Plotting only available for rainfall, wind, waterlevel, and discharge forcings."
            )


def plot_discharge(
    event: Event,
    site: Site,
) -> tuple[str, Optional[List[Exception]]]:
    rivers: List[IDischarge] = event.forcings.get(ForcingType.DISCHARGE)
    if site.sfincs.river is None:
        raise ValueError("No rivers defined for this site.")
    elif not rivers:
        return "", None
    logger.debug("Plotting discharge data")

    units = site.gui.units.default_discharge_units

    data = pd.DataFrame()
    errors = []

    for discharge in rivers:
        try:
            if discharge.source in UNPLOTTABLE_SOURCES:
                logger.debug(
                    f"Plotting not supported for discharge data from `{discharge.source}`"
                )
                continue
            elif isinstance(
                discharge, (DischargeConstant, DischargeSynthetic, DischargeCSV)
            ):
                river_data = discharge.to_dataframe(event.time)
            else:
                raise ValueError(f"Unknown discharge source: `{discharge.source}`")

            # Rename columns to avoid conflicts
            river_data.columns = [discharge.river.name]
            if data.empty:
                data = river_data
            else:
                # add river_data as a column to the dataframe. keep the same index
                data = data.join(river_data, how="outer")
        except Exception as e:
            errors.append((discharge.river.name, e))

    if errors:
        logger.error(
            f"Could not retrieve discharge data for {', '.join([entry[0] for entry in errors])}: {errors}"
        )
        return "", errors

    river_names, river_descriptions = [], []
    for river in site.sfincs.river:
        river_names.append(river.name)
        river_descriptions.append(river.description or river.name)

    if event.template == Template.Synthetic:
        data.index = (
            data.index - data.index[0]
        ).total_seconds() / 3600  # Convert to hours
        x_title = "Hours from start"
    else:
        x_title = "Time"

    # Plot actual thing
    fig = go.Figure()
    for ii, col in enumerate(data.columns):
        fig.add_trace(
            go.Scatter(
                x=data.index,
                y=data[col],
                name=river_descriptions[ii],
                mode="lines",
            )
        )

    fig.update_layout(
        autosize=False,
        height=100 * 2,
        width=280 * 2,
        margin={"r": 0, "l": 0, "b": 0, "t": 0},
        font={"size": 10, "color": "black", "family": "Arial"},
        title_font={"size": 10, "color": "black", "family": "Arial"},
        yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        xaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        xaxis_title={"text": x_title},
        yaxis_title={"text": f"River discharge [{units.value}]"},
        xaxis={"range": [event.time.start_time, event.time.end_time]},
    )

    # Only save to the the event folder if that has been created already.
    # Otherwise this will create the folder and break the db since there is no event.toml yet
    output_dir = db_path(object_dir="events", obj_name=event.name)
    if not output_dir.exists():
        output_dir = gettempdir()
    output_loc = Path(output_dir) / "discharge_timeseries.html"
    if output_loc.exists():
        output_loc.unlink()
    fig.write_html(output_loc)
    return output_loc.as_posix(), None


def plot_waterlevel(
    event: Event,
    site: Site,
) -> tuple[str, Optional[List[Exception]]]:
    forcing_list = event.forcings.get(ForcingType.WATERLEVEL)
    if not forcing_list:
        return "", None
    elif site.sfincs.water_level is None:
        raise ValueError("No water levels defined for this site.")

    waterlevel = forcing_list[0]
    if waterlevel.source in UNPLOTTABLE_SOURCES:
        logger.debug(
            f"Plotting not supported for waterlevel data from {waterlevel.source}"
        )
        return "", None

    logger.debug("Plotting water level data")
    units = site.gui.units.default_length_units
    data = None
    try:
        if isinstance(waterlevel, WaterlevelGauged):
            if site.sfincs.tide_gauge is None:
                raise ValueError("No tide gauge defined for this site.")
            data = site.sfincs.tide_gauge.get_waterlevels_in_time_frame(
                event.time, units=units
            )

            # Convert to main reference
            datum_correction = site.sfincs.water_level.get_datum(
                site.sfincs.tide_gauge.reference
            ).height.convert(units)
            data += datum_correction

        elif isinstance(waterlevel, WaterlevelCSV):
            data = waterlevel.to_dataframe(event.time)
        elif isinstance(waterlevel, WaterlevelSynthetic):
            data = waterlevel.to_dataframe(time_frame=event.time)
            datum_correction = site.sfincs.water_level.get_datum(
                site.gui.plotting.synthetic_tide.datum
            ).height.convert(units)
            data += datum_correction
        else:
            raise ValueError(f"Unknown waterlevel type: {waterlevel}")

    except Exception as e:
        logger.error(f"Error getting water level data: {e}")
        return "", [e]

    if data is not None and data.empty:
        logger.error(f"Could not retrieve waterlevel data: {waterlevel} {data}")
        return "", None

    if event.template == Template.Synthetic:
        data.index = (
            data.index - data.index[0]
        ).total_seconds() / 3600  # Convert to hours
        x_title = "Hours from start"
    else:
        x_title = "Time"

    # Plot actual thing
    fig = px.line(data)

    # plot main reference
    fig.add_hline(
        y=0,
        line_dash="dash",
        line_color="#000000",
        annotation_text=site.sfincs.water_level.reference,
        annotation_position="bottom right",
    )

    # plot other references
    for wl_ref in site.sfincs.water_level.datums:
        if (
            wl_ref.name == site.sfincs.config.overland_model.reference
            or wl_ref.name in site.gui.plotting.excluded_datums
        ):
            continue

        fig.add_hline(
            y=wl_ref.height.convert(units),
            line_dash="dash",
            line_color="#3ec97c",
            annotation_text=wl_ref.name,
            annotation_position="bottom right",
        )

    fig.update_layout(
        autosize=False,
        height=100 * 2,
        width=280 * 2,
        margin={"r": 0, "l": 0, "b": 0, "t": 0},
        font={"size": 10, "color": "black", "family": "Arial"},
        title_font={"size": 10, "color": "black", "family": "Arial"},
        legend=None,
        xaxis_title=x_title,
        yaxis_title=f"Water level [{units.value}]",
        yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        xaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        showlegend=False,
        xaxis={"range": [data.index.min(), data.index.max()]},
    )

    # Only save to the the event folder if that has been created already.
    # Otherwise this will create the folder and break the db since there is no event.toml yet
    output_dir = db_path(object_dir="events", obj_name=event.name)
    if not output_dir.exists():
        output_dir = gettempdir()
    output_loc = Path(output_dir) / "waterlevel_timeseries.html"
    if output_loc.exists():
        output_loc.unlink()
    fig.write_html(output_loc)
    return output_loc.as_posix(), None


def plot_rainfall(
    event: Event,
    site: Site,
) -> tuple[str, Optional[List[Exception]]]:
    forcing_list = event.forcings.get(ForcingType.RAINFALL)
    if not forcing_list:
        return "", None
    elif forcing_list[0].source in UNPLOTTABLE_SOURCES:
        logger.warning(
            f"Plotting not supported for rainfall datafrom sources {', '.join(UNPLOTTABLE_SOURCES)}"
        )
        return "", None

    rainfall = forcing_list[0]
    logger.debug("Plotting rainfall data")

    data = None
    try:
        if isinstance(rainfall, (RainfallConstant, RainfallCSV, RainfallSynthetic)):
            data = rainfall.to_dataframe(event.time)
        else:
            raise ValueError(f"Unknown rainfall type: {rainfall}")
    except Exception as e:
        logger.error(f"Error getting rainfall data: {e}")
        return "", [e]

    if data is None or data.empty:
        logger.error(f"Could not retrieve rainfall data: {rainfall} {data}")
        return "", None

    # Add multiplier
    data *= event.rainfall_multiplier

    if event.template == Template.Synthetic:
        data.index = (
            data.index - data.index[0]
        ).total_seconds() / 3600  # Convert to hours
        x_title = "Hours from start"
    else:
        x_title = "Time"

    # Plot actual thing
    fig = px.line(data_frame=data)

    fig.update_layout(
        autosize=False,
        height=100 * 2,
        width=280 * 2,
        margin={"r": 0, "l": 0, "b": 0, "t": 0},
        font={"size": 10, "color": "black", "family": "Arial"},
        title_font={"size": 10, "color": "black", "family": "Arial"},
        legend=None,
        yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        xaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        xaxis_title={"text": x_title},
        yaxis_title={
            "text": f"Rainfall intensity [{site.gui.units.default_intensity_units.value}]"
        },
        showlegend=False,
        xaxis={"range": [event.time.start_time, event.time.end_time]},
    )
    # Only save to the the event folder if that has been created already.
    # Otherwise this will create the folder and break the db since there is no event.toml yet
    output_dir = db_path(object_dir="events", obj_name=event.name)
    if not output_dir.exists():
        output_dir = gettempdir()
    output_loc = Path(output_dir) / "rainfall_timeseries.html"
    if output_loc.exists():
        output_loc.unlink()
    fig.write_html(output_loc)
    return output_loc.as_posix(), None


def plot_wind(
    event: Event,
    site: Site,
) -> tuple[str, Optional[List[Exception]]]:
    logger.debug("Plotting wind data")
    forcing_list = event.forcings.get(ForcingType.WIND)
    if not forcing_list:
        return "", None
    elif forcing_list[0].source in UNPLOTTABLE_SOURCES:
        logger.warning(
            f"Plotting not supported for wind data from sources {', '.join(UNPLOTTABLE_SOURCES)}"
        )
        return "", None

    wind = forcing_list[0]
    data = None
    try:
        if isinstance(wind, (WindConstant, WindCSV, WindSynthetic)):
            data = wind.to_dataframe(event.time)
        else:
            raise ValueError(f"Unknown wind type: {wind}")
    except Exception as e:
        logger.error(f"Error getting wind data: {e}")
        return "", [e]

    if data is None or data.empty:
        logger.error(
            f"Could not retrieve wind data: {event.forcings.get(ForcingType.WIND)} {data}"
        )
        return "", None

    if event.template == Template.Synthetic:
        data.index = (
            data.index - data.index[0]
        ).total_seconds() / 3600  # Convert to hours
        x_title = "Hours from start"
    else:
        x_title = "Time"

    # Plot actual thing
    fig = px.line(x=data.index, y=data.iloc[:, 0])

    fig.update_layout(
        autosize=False,
        height=100 * 2,
        width=280 * 2,
        margin={"r": 0, "l": 0, "b": 0, "t": 0},
        font={"size": 10, "color": "black", "family": "Arial"},
        title_font={"size": 10, "color": "black", "family": "Arial"},
        legend=None,
        showlegend=False,
        xaxis={"range": [event.time.start_time, event.time.end_time]},
        xaxis_title={"text": x_title},
        xaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
        yaxis_title={
            "text": f"Wind speed [{site.gui.units.default_velocity_units.value}]"
        },
    )

    # Only save to the the event folder if that has been created already.
    # Otherwise this will create the folder and break the db since there is no event.toml yet
    output_dir = db_path(object_dir="events", obj_name=event.name)
    if not output_dir.exists():
        output_dir = gettempdir()
    output_loc = Path(output_dir) / "wind_timeseries.html"
    if output_loc.exists():
        output_loc.unlink()
    fig.write_html(output_loc)
    return output_loc.as_posix(), None



from pathlib import Path

from flood_adapt.dbs_classes.dbs_template import DbsTemplate
from flood_adapt.misc.exceptions import DoesNotExistError
from flood_adapt.objects.events.event_factory import EventFactory
from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.events.events import Event


class DbsEvent(DbsTemplate[Event]):
    dir_name = "events"
    display_name = "Event"
    _object_class = Event
    _higher_lvl_object = "Scenario"

    def get(self, name: str, load_all: bool = False) -> Event | EventSet:
        """Return an event object.

        Parameters
        ----------
        name : str
            name of the event to be returned

        Returns
        -------
        Event
            event object
        """
        # Get event path
        event_path = self.input_path / f"{name}" / f"{name}.toml"

        # Check if the object exists
        if not Path(event_path).is_file():
            raise DoesNotExistError(name, self.display_name)

        # Load event
        return EventFactory.load_file(event_path, load_all=load_all)

    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if an event is used in a scenario.

        Parameters
        ----------
        name : str
            name of the event to be checked

        Returns
        -------
        list[str]
            list of scenarios that use the event
        """
        # Get all the scenarios
        scenarios = [
            self._database.scenarios.get(scn)
            for scn in self._database.scenarios.summarize_objects()["name"]
        ]

        # Check if event is used in a scenario
        used_in_scenario = [
            scenario.name for scenario in scenarios if name == scenario.event
        ]
        return used_in_scenario



from flood_adapt.objects.measures.measures import (
    HazardMeasure,
    ImpactMeasure,
    Measure,
)
from flood_adapt.objects.object_model import Object


class Strategy(Object):
    """
    Class representing a strategy in FloodAdapt.

    A strategy is a collection of measures that can be applied to a model.

    Attributes
    ----------
    measures : list[str]
        A list of measures associated with the strategy. Should be a list of measure names that are saved in the database.

    """

    measures: list[str]

    _measure_objects: list[Measure] | None = None

    def initialize_measure_objects(self, measures: list[Measure]) -> None:
        """Initialize the measure objects associated with this strategy.

        Parameters
        ----------
        measures : list[Measure]
            A list of measure objects to be associated with this strategy. Should be a list of measure objects that are saved in the database.
        """
        self._measure_objects = measures

    def get_measures(self) -> list[Measure]:
        """Get the measures associated with this strategy.

        Note that this method will return the measure objects, not just their names.
        The measure objects are initialized using the `initialize_measure_objects` method.

        Returns
        -------
        measures : list[Measure]
            The list of measure objects associated with this strategy.

        Raises
        ------
        ValueError
            If the measure objects have not been initialized.
        """
        # Get measure paths using a database structure
        if self._measure_objects is None:
            raise ValueError(
                "Measure objects have not been initialized. Call `initialize_measure_objects()` first."
            )
        return self._measure_objects

    def get_impact_measures(self) -> list[ImpactMeasure]:
        return [
            measure
            for measure in self.get_measures()
            if isinstance(measure, ImpactMeasure)
        ]

    def get_impact_strategy(self) -> "Strategy":
        impact_measures = self.get_impact_measures()
        impact_strategy = Strategy(
            name=self.name,
            measures=[m.name for m in impact_measures],
        )
        impact_strategy.initialize_measure_objects(impact_measures)
        return impact_strategy

    def get_hazard_measures(self) -> list[HazardMeasure]:
        return [
            measure
            for measure in self.get_measures()
            if isinstance(measure, HazardMeasure)
        ]

    def get_hazard_strategy(self) -> "Strategy":
        hazard_measures = self.get_hazard_measures()
        hazard_strategy = Strategy(
            name=self.name,
            measures=[m.name for m in hazard_measures],
        )
        hazard_strategy.initialize_measure_objects(hazard_measures)
        return hazard_strategy



from flood_adapt.dbs_classes.database import Database
from flood_adapt.dbs_classes.dbs_benefit import DbsBenefit
from flood_adapt.dbs_classes.dbs_event import DbsEvent
from flood_adapt.dbs_classes.dbs_measure import DbsMeasure
from flood_adapt.dbs_classes.dbs_projection import DbsProjection
from flood_adapt.dbs_classes.dbs_scenario import DbsScenario
from flood_adapt.dbs_classes.dbs_static import DbsStatic
from flood_adapt.dbs_classes.dbs_strategy import DbsStrategy
from flood_adapt.dbs_classes.dbs_template import DbsTemplate

__all__ = [
    "Database",
    "DbsBenefit",
    "DbsEvent",
    "DbsMeasure",
    "DbsProjection",
    "DbsScenario",
    "DbsStatic",
    "DbsStrategy",
    "DbsTemplate",
]
